#!/usr/bin/env python
#
# Copyright (c), 2021, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
# Authors: Davide Brunato
#
import unittest
import numpy as np
from pathlib import Path

from qeschema.upf import read_pseudo_file


class TestUpfUtils(unittest.TestCase):

    def test_read_pseudo_file(self):
        upf_file = Path(__file__).parent.joinpath('resources/upf/N.pz-vbc.UPF')

        obj = read_pseudo_file(str(upf_file))
        self.assertIsInstance(obj, dict)
        self.assertListEqual(list(obj), ['PP_INFO', 'PP_HEADER', 'PP_MESH',
                                         'PP_LOCAL', 'PP_RHOATOM', 'PP_NONLOCAL'])

        self.assertIsInstance(obj['PP_INFO'], dict)
        self.assertIsInstance(obj['PP_HEADER'], dict)
        self.assertIsInstance(obj['PP_MESH'], dict)
        self.assertIsInstance(obj['PP_MESH']['PP_R'], np.ndarray)
        self.assertIsInstance(obj['PP_LOCAL'], np.ndarray)
        self.assertIsInstance(obj['PP_RHOATOM'], np.ndarray)
        self.assertIsInstance(obj['PP_NONLOCAL'], dict)


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c), 2015-2019, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
# Authors: Davide Brunato
#
import unittest
import logging
import re

from qeschema.cards import get_atomic_species_card, get_atomic_positions_cell_card, \
    get_atomic_constraints_card, get_k_points_card, get_atomic_forces_card, \
    get_cell_parameters_card, get_qpoints_card, get_climbing_images, \
    get_neb_images_positions_card, get_neb_cell_parameters_card, \
    get_neb_atomic_forces_card, get_positions_units

logger = logging.getLogger('qeschema')


class TestCardsFunctions(unittest.TestCase):

    def test_get_atomic_species_card(self):
        kwargs = {
            'atomic_species': {
                'species': [{'@name': 'Al', 'mass': 26.981538, 'pseudo_file': 'Al.pbe-n-van.UPF'},
                            {'@name': 'H', 'mass': 1.00794, 'pseudo_file': 'H.pbe-van_ak.UPF'}]
            }}

        result = get_atomic_species_card('ATOMIC_SPECIES', **kwargs)
        self.assertListEqual(result, ['ATOMIC_SPECIES',
                                      ' Al 26.981538 Al.pbe-n-van.UPF',
                                      ' H 1.00794 H.pbe-van_ak.UPF'])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_atomic_species_card('ATOMIC_SPECIES')

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments when "
                                              "building ATOMIC_SPECIES card! 'atomic_species'"])

    def test_get_positions_units(self):
        item = {
            '@nat': 2,
            '@alat': 1.0,
            'crystal_positions': {'atom': [
                {'@name': 'Fe', '@index': 1, '$': [0.0, 0.0, 0.0]},
                {'@name': 'Fe1', '@index': 2, '$': [0.5, 0.5, 0.5]}
            ]},
            'cell': {'a1': [5.406506, 0.0, 0.0],
                     'a2': [0.0, 5.406506, 0.0],
                     'a3': [0.0, 0.0, 5.406506]}
        }

        result, unit = get_positions_units(item)
        self.assertDictEqual(result, item['crystal_positions'])
        self.assertEqual(unit, 'crystal')

        result, unit = get_positions_units(item={})
        self.assertDictEqual(result, {})
        self.assertIsNone(unit)

    def test_get_atomic_positions_cell_card(self):
        kwargs = {
            'atomic_structure': {
                '@nat': 4,
                '@alat': 8.19,
                'atomic_positions': {
                    'atom': [{'@name': 'O1', '$': [0.5, 0.5, 0.5]},
                             {'@name': 'O1', '$': [1.5, 1.5, 1.5]},
                             {'@name': 'Fe1', '$': [0.0, 0.0, 0.0]},
                             {'@name': 'Fe2', '$': [1.0, 1.0, 1.0]}]
                },
                'cell': {'a1': [0.5, 0.5, 1.0], 'a2': [0.5, 1.0, 0.5], 'a3': [1.0, 0.5, 0.5]}
            },
        }

        result = get_atomic_positions_cell_card('ATOMIC_POSITIONS', **kwargs)
        self.assertListEqual([re.sub(r'\s+', ' ', s) for s in result],
                             ['ATOMIC_POSITIONS bohr',
                              'O1 0.50000000 0.50000000 0.50000000',
                              'O1 1.50000000 1.50000000 1.50000000',
                              'Fe1 0.00000000 0.00000000 0.00000000',
                              'Fe2 1.00000000 1.00000000 1.00000000'])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_atomic_positions_cell_card('ATOMIC_POSITIONS')

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema:Missing required arguments '
                                          'when building ATOMIC_POSITIONS card!'])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_atomic_positions_cell_card(
                'ATOMIC_POSITIONS', atomic_structure={'atomic_positions': {}}
            )
        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema:Cannot find any atoms '
                                          'for building ATOMIC_POSITIONS!'])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_atomic_positions_cell_card(
                'ATOMIC_POSITIONS', atomic_structure={'atomic_positions': {'something': True}}
            )
        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema:Cannot find any atoms '
                                          'for building ATOMIC_POSITIONS!'])

        # Test with wyckoff_positions
        kwargs = {
            'atomic_structure': {
                '@nat': 1,
                '@alat': 7.42,
                'wyckoff_positions': {'atom': {'@name': 'Pt', '$': [0.0, 0.0, 0.0]}},
                'cell': {'a1': [-0.5, 0.0, 0.5],
                         'a2': [0.0, 0.5, 0.5],
                         'a3': [-0.5, 0.5, 0.0]}
            }
        }
        result = get_atomic_positions_cell_card('ATOMIC_POSITIONS', **kwargs)
        self.assertListEqual(result, ['ATOMIC_POSITIONS crystal_sg',
                                      'Pt     0.00000000    0.00000000    0.00000000'])

        # Test with logging errors
        kwargs = {
            'atomic_structure': {
                '@nat': 2,
                '@alat': 12.0,
                'atomic_positions': {'atom': [{'@name': 'C', '$': ['0.188', 0.0, 0.0]},
                                              {'@name': 'O', '$': [0.0, 0.0, 0.0]}]},
                'cell': {'a1': [1.0, 0.0, 0.0], 'a2': [0.0, 1.0, 0.0], 'a3': [0.0, 0.0, 1.0]}
            },
            'free_positions': {'@rank': 1, '@dims': [1], '$': [1]}
        }

        with self.assertLogs('qeschema', 'ERROR') as ctx:
            result = get_atomic_positions_cell_card('ATOMIC_POSITIONS', **kwargs)
        self.assertEqual(ctx.output, [
            'ERROR:qeschema:ATOMIC_POSITIONS: incorrect number of position constraints!',
            'ERROR:qeschema:ATOMIC_POSITIONS: incorrect datatype in positions!',
        ])
        self.assertListEqual(result, ['ATOMIC_POSITIONS bohr',
                                      'O      0.00000000    0.00000000    0.00000000'])

    def test_get_atomic_constraints_card(self):
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_atomic_constraints_card('CONSTRAINTS')

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema:Missing required arguments '
                                          'when building CONSTRAINTS card!'])

        result = get_atomic_constraints_card(
                'CONSTRAINTS', **{'atomic_constraints': {'num_of_constraints': 1,
                                                         'tolerance': 0.2,
                                                         'atomic_constraint': []}})
        self.assertListEqual(result, ['CONSTRAINTS', '1 0.2'])
        result = get_atomic_constraints_card(
                'CONSTRAINTS',
                **{'atomic_constraints': {'num_of_constraints': 1, 'tolerance': 0.2,
                                          'atomic_constraint': [
                                             {'constr_parms': [0.2, 0.3, 0.1, 0.9],
                                              'constr_type': 'test_constraint',
                                              'constr_target': 0.3
                                              }]
                                          }
                   })
        self.assertListEqual(result, ['CONSTRAINTS', '1 0.2',
                                      'test_constraint 0.2 0.3 0.1 0.9 0.3'])

    def test_get_k_points_card(self):
        result = get_k_points_card('K_POINTS', k_points_IBZ={
            'monkhorst_pack': {'@nk1': 6, '@nk2': 6, '@nk3': 1, '@k1': 1, '@k2': 1, '@k3': 1}
        })
        self.assertListEqual(result, ['K_POINTS automatic', ' 6 6 1 1 1 1'])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_k_points_card('K_POINTS')

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments "
                                          "when building K_POINTS card! 'k_points_IBZ'"])

    def test_get_atomic_forces_card(self):
        kwargs = {
            'atomic_positions': {'atom': [{'@name': 'H', '@index': 1,
                                           '$': [-4.56670009, 0.0, 0.0]},
                                          {'@name': 'H', '@index': 2,
                                           '$': [0.0, 0.0, 0.0]},
                                          {'@name': 'H', '@index': 3,
                                           '$': [1.55776676, 0.0, 0.0]},
                                          {'@name': 'H', '@index': 1,
                                           '$': [-1.55776676, 0.0, 0.0]},
                                          {'@name': 'H', '@index': 2,
                                           '$': [0.0, 0.0, 0.0]},
                                          {'@name': 'H', '@index': 3,
                                           '$': [4.56670009, 0.0, 0.0]}]},
            'external_atomic_forces': [0.0, 0.0, 0.0,
                                       0.0, 0.0, 0.1,
                                       0.0, 0.0, -0.1,
                                       0.2, 0.0, 0.0,
                                       -0.2, 0.0, 0.0,
                                       0.0, 0.0, 0.0]
        }
        result = get_atomic_forces_card('ATOMIC_FORCES', **kwargs)
        self.assertListEqual(result, ['ATOMIC_FORCES', 
                                      'H        0.00000000     0.00000000     0.00000000',
                                      'H        0.00000000     0.00000000     0.10000000',
                                      'H        0.00000000     0.00000000    -0.10000000',
                                      'H        0.20000000     0.00000000     0.00000000',
                                      'H       -0.20000000     0.00000000     0.00000000',
                                      'H        0.00000000     0.00000000     0.00000000'])

        kwargs['external_atomic_forces'].pop()
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_atomic_forces_card('ATOMIC_FORCES', **kwargs)
        self.assertListEqual(result, ['ATOMIC_FORCES',
                                      'H        0.00000000     0.00000000     0.00000000',
                                      'H        0.00000000     0.00000000     0.10000000',
                                      'H        0.00000000     0.00000000    -0.10000000',
                                      'H        0.20000000     0.00000000     0.00000000',
                                      'H       -0.20000000     0.00000000     0.00000000'])
        self.assertListEqual(context.output, ['ERROR:qeschema:incorrect number of atomic forces'])

        # with self.assertLogs(logger, level='DEBUG') as context:
        #    result = get_atomic_forces_card('ATOMIC_FORCES', **kwargs)
        # self.assertListEqual(result,[])
        # self.assertEqual(context.output, ['DEBUG:qeschema:Missing required arguments '
        #                                  'when building ATOMIC_FORCES card!'])

        # FIXME: a fix is required in get_atomic_forces_card()
        # result = get_atomic_forces_card('ATOMIC_FORCES',
        #                                 external_atomic_forces=[0.1, 0.2], **kwargs)
        # self.assertListEqual(result, [])

    def test_get_cell_parameters_card(self):
        kwargs = {
            'atomic_structure': {
                '@nat': 1,
                '@alat': 7.5,
                'atomic_positions': {'atom': [{'@name': 'Al', '$': [0.0, 0.0, 0.0]}]},
                'cell': {'a1': [-0.5, 0.0, 0.5],
                         'a2': [0.0, 0.5, 0.5],
                         'a3': [-0.5, 0.5, 0.0],
                         'a4': 'test only, this item is skipped.'}},
        }

        result = get_cell_parameters_card('CELL_PARAMETERS', **kwargs)
        self.assertListEqual([re.sub(r'\s+', ' ', s) for s in result],
                             ['CELL_PARAMETERS bohr',
                              ' -0.50000000 0.00000000 0.50000000 ',
                              ' 0.00000000 0.50000000 0.50000000 ',
                              ' -0.50000000 0.50000000 0.00000000 '])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_cell_parameters_card('CELL_PARAMETERS')

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema:Missing required arguments '
                                          'when building CELL_PARAMETERS card!'])

        kwargs['atomic_structure'].pop('cell')
        result = get_cell_parameters_card('CELL_PARAMETERS', **kwargs)
        self.assertListEqual(result, [])

    def test_get_qpoints_card(self):
        result = get_qpoints_card('qPointsSpecs', ldisp=True)
        self.assertListEqual(result, [])

        result = get_qpoints_card('qPointsSpecs', ldisp=False)
        self.assertListEqual([re.sub(r'\s+', ' ', s) for s in result],
                             ['0.0000 0.0000 0.0000'])

        kwargs = {
            'qplot': True,
            'nqs': 10,
            'q_points_list': {
                'q_point': [{'@weight': 0.1, '$': [0.1, 0.0, 0.0]},
                            {'@weight': 0.1, '$': [0.5, 0.0, 0.0]},
                            {'@weight': 0.1, '$': [1.0, 0.0, 0.0]}]
            }
        }
        result = get_qpoints_card('qPointsSpecs', **kwargs)
        self.assertListEqual([re.sub(r'\s+', ' ', s) for s in result],
                             [' 10', ' 0.1 0.0 0.0 0.1', ' 0.5 0.0 0.0 0.1', ' 1.0 0.0 0.0 0.1'])

        kwargs.pop('nqs')
        with self.assertLogs(logger, level='WARNING') as context:
            result = get_qpoints_card('qPointsSpecs', **kwargs)
        self.assertListEqual(context.output, ['WARNING:qeschema:qplot was set to true in input '
                                              'but no value for nqs was provided: assuming nqs=1'])
        self.assertListEqual([re.sub(r'\s+', ' ', s) for s in result],
                             [' 1', ' 0.1 0.0 0.0 0.1', ' 0.5 0.0 0.0 0.1', ' 1.0 0.0 0.0 0.1'])

    def test_get_climbing_images(self):
        result = get_climbing_images('CLIMBING_IMAGES', climbingImage='no-CI')
        self.assertListEqual(result, [''])

        result = get_climbing_images('CLIMBING_IMAGES', climbingImage='auto')
        self.assertListEqual(result, [''])

        result = get_climbing_images('CLIMBING_IMAGES', climbingImage='manual',
                                     climbingImageIndex=[])
        self.assertListEqual(result, [''])

        result = get_climbing_images('CLIMBING_IMAGES', climbingImage='manual',
                                     climbingImageIndex=10)
        self.assertListEqual(result, [' 10 '])

        result = get_climbing_images('CLIMBING_IMAGES')
        self.assertListEqual(result, [''])

    def test_get_neb_images_positions_card(self):
        kwargs = {
            'atomic_structure': [
                {
                    '@nat': 3,
                    '@alat': 12.0,
                    'atomic_positions': {
                        'atom': [{'@name': 'H', '@index': 1, '$': [-4.56670009, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 2, '$': [0.0, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 3, '$': [1.55776676, 0.0, 0.0]}]},
                    'cell': {'a1': [12.0, 0.0, 0.0],
                             'a2': [0.0, 12.0, 0.0],
                             'a3': [0.0, 0.0, 12.0]}},
                {
                    '@nat': 3,
                    '@alat': 12.0,
                    'atomic_positions': {
                        'atom': [{'@name': 'H', '@index': 1, '$': [-1.55776676, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 2, '$': [0.0, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 3, '$': [4.56670009, 0.0, 0.0]}]},
                    'cell': {'a1': [12.0, 0.0, 0.0],
                             'a2': [0.0, 12.0, 0.0],
                             'a3': [0.0, 0.0, 12.0]}}
            ]}

        result = get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertEqual(len(result), 12)
        self.assertListEqual(result, [
            'BEGIN_POSITIONS ',
            'FIRST_IMAGE ',
            'ATOMIC_POSITIONS { bohr }',
            'H     -4.56670009    0.00000000    0.00000000',
            'H      0.00000000    0.00000000    0.00000000',
            'H      1.55776676    0.00000000    0.00000000',
            'LAST_IMAGE ',
            'ATOMIC_POSITIONS { bohr }',
            'H     -1.55776676    0.00000000    0.00000000',
            'H      0.00000000    0.00000000    0.00000000',
            'H      4.56670009    0.00000000    0.00000000',
            'END_POSITIONS '
        ])

        kwargs['free_positions'] = {
            '@rank': 2, '@dims': [3, 3], '$': [1, 1]  # Wrong number of free positions
        }
        with self.assertLogs(logger, level='ERROR') as ctx:
            get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertListEqual(ctx.output, ['ERROR:qeschema:ATOMIC_POSITIONS: '
                                          'incorrect number of position constraints!'])

        kwargs['free_positions'] = {
            '@rank': 2, '@dims': [3, 3], '$': [1, 1, 1, 1, 1, 1, 1, 1, 1]
        }
        result = get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertEqual(len(result), 12)
        self.assertListEqual(result, [
            'BEGIN_POSITIONS ',
            'FIRST_IMAGE ',
            'ATOMIC_POSITIONS { bohr }',
            'H     -4.56670009    0.00000000    0.00000000',
            'H      0.00000000    0.00000000    0.00000000',
            'H      1.55776676    0.00000000    0.00000000',
            'LAST_IMAGE ',
            'ATOMIC_POSITIONS { bohr }',
            'H     -1.55776676    0.00000000    0.00000000',
            'H      0.00000000    0.00000000    0.00000000',
            'H      4.56670009    0.00000000    0.00000000',
            'END_POSITIONS '
        ])
        kwargs.pop('free_positions')

        kwargs['atomic_structure'].append(kwargs['atomic_structure'][-1])
        result = get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertEqual(len(result), 17)
        self.assertListEqual(result, [
            'BEGIN_POSITIONS ',
            'FIRST_IMAGE ',
            'ATOMIC_POSITIONS { bohr }',
            'H     -4.56670009    0.00000000    0.00000000',
            'H      0.00000000    0.00000000    0.00000000',
            'H      1.55776676    0.00000000    0.00000000',
            'INTERMEDIATE_IMAGE ',
            'ATOMIC_POSITIONS { bohr }',
            'H     -1.55776676    0.00000000    0.00000000',
            'H      0.00000000    0.00000000    0.00000000',
            'H      4.56670009    0.00000000    0.00000000',
            'LAST_IMAGE ',
            'ATOMIC_POSITIONS { bohr }',
            'H     -1.55776676    0.00000000    0.00000000',
            'H      0.00000000    0.00000000    0.00000000',
            'H      4.56670009    0.00000000    0.00000000',
            'END_POSITIONS '
        ])

        kwargs['atomic_structure'][-1]['atomic_positions']['atom'].pop()
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertEqual(context.output[0],
                         'ERROR:qeschema:found images with differing number of atoms !!!')
        self.assertEqual(result, '')

        kwargs['atomic_structure'] = kwargs['atomic_structure'][0]
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertListEqual(context.output, ['ERROR:qeschema:at least the atomic structures '
                                              'for first and last image should be provided'])
        self.assertListEqual(result, [])

        kwargs = {
            'atomic_structure': [
                {
                    '@nat': 3,
                    '@alat': 12.0,
                    'atomic_positions': {
                        'atom': [{'@name': 'H', '@index': 1, '$': [-4.56670009, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 2, '$': [0.0, 0.0, 0.0]}]},
                    'cell': {'a1': [12.0, 0.0, 0.0],
                             'a2': [0.0, 12.0, 0.0],
                             'a3': [0.0, 0.0, 12.0]}},
                {
                    '@nat': 3,
                    '@alat': 12.0,
                    'atomic_positions': {
                        'atom': [{'@name': 'H', '@index': 1, '$': [-1.55776676, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 2, '$': [0.0, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 3, '$': [4.56670009, 0.0, 0.0]}]},
                    'cell': {'a1': [12.0, 0.0, 0.0],
                             'a2': [0.0, 12.0, 0.0],
                             'a3': [0.0, 0.0, 12.0]}}
            ]}

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertListEqual(context.output, [
            'ERROR:qeschema:nat provided in first image differs from number '
            'of atoms in atomic_positions!!!'])
        self.assertEqual(result, '')

        kwargs['free_positions'] = {'@rank': 1, '@dims': [1], '$': [1]}
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertListEqual(context.output, [
            'ERROR:qeschema:nat provided in first image differs from number '
            'of atoms in atomic_positions!!!'])
        self.assertEqual(result, '')

        kwargs['atomic_structure'][0]['atomic_positions']['atom'].clear()
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_neb_images_positions_card('ATOMIC_POSITIONS', **kwargs)
        self.assertListEqual(context.output, [
            'ERROR:qeschema:no atomic coordinates provided for first image'
        ])
        self.assertEqual(result, '')

    def test_get_neb_cell_parameters_card(self):
        kwargs = {
            'atomic_structure': [
                {
                    '@nat': 3,
                    '@alat': 12.0,
                    'crystal_positions': {
                        'atom': [{'@name': 'H', '@index': 1, '$': [-0.38055834, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 2, '$': [0.0, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 3, '$': [0.1298139, 0.0, 0.0]}]},
                    'cell': {'a0': [12.0, 0.0, 0.0],  # Ignored, only for testing
                             'a1': [12.0, 0.0, 0.0],
                             'a2': [0.0, 12.0, 0.0],
                             'a3': [0.0, 0.0, 12.0]}},
                {
                    '@nat': 3,
                    '@alat': 12.0,
                    'crystal_positions': {
                        'atom': [{'@name': 'H', '@index': 1, '$': [-0.1298139, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 2, '$': [0.0, 0.0, 0.0]},
                                 {'@name': 'H', '@index': 3, '$': [0.38055834, 0.0, 0.0]}]},
                    'cell': {'a1': [12.0, 0.0, 0.0],
                             'a2': [0.0, 12.0, 0.0],
                             'a3': [0.0, 0.0, 12.0]}}]}

        result = get_neb_cell_parameters_card('CELL_PARAMETERS', **kwargs)
        self.assertListEqual([re.sub(r'\s+', ' ', s) for s in result],
                             ['CELL_PARAMETERS bohr',
                              ' 12.00000000 0.00000000 0.00000000 ',
                              ' 0.00000000 12.00000000 0.00000000 ',
                              ' 0.00000000 0.00000000 12.00000000 '])

        result = get_neb_cell_parameters_card('CELL_PARAMETERS',
                                              atomic_structure=kwargs['atomic_structure'][0])
        self.assertListEqual([re.sub(r'\s+', ' ', s) for s in result],
                             ['CELL_PARAMETERS bohr',
                              ' 12.00000000 0.00000000 0.00000000 ',
                              ' 0.00000000 12.00000000 0.00000000 ',
                              ' 0.00000000 0.00000000 12.00000000 '])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_neb_cell_parameters_card('CELL_PARAMETERS', atomic_structure=[])

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema: No atomic_structure '
                                          'element found in kwargs !!!'])

        kwargs['atomic_structure'][0]['cell'].clear()
        result = get_neb_cell_parameters_card('CELL_PARAMETERS', **kwargs)
        self.assertListEqual(result, [])

    def test_get_neb_atomic_forces_card(self):
        self.assertIsNone(get_neb_atomic_forces_card('atomic_forces'))


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
#
# Copyright (c), 2021, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
# Authors: Davide Brunato
#
import unittest
import numpy as np
from pathlib import Path

try:
    import yaml
except ImportError:
    yaml = None
else:
    from qeschema.hdf5 import read_charge_file, get_wavefunctions, \
        get_wf_attributes, get_wfc_miller_indices


    # TODO: Fetch appropriate HDF5 files for testing

    @unittest.SkipTest
    class TestHdf5Utils(unittest.TestCase):

        def test_read_charge_file(self):
            hdf5_file = Path(__file__).parent.joinpath('resources/hdf5/???')

        def test_get_wf_attributes(self):
            hdf5_file = Path(__file__).parent.joinpath('resources/hdf5/???')

        def test_get_wavefunctions(self):
            hdf5_file = Path(__file__).parent.joinpath('resources/hdf5/???')

        def test_get_wfc_miller_indices(self):
            hdf5_file = Path(__file__).parent.joinpath('resources/hdf5/???')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c), 2015-2019, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
# Authors: Davide Brunato
#
import unittest
import tempfile
import logging
import sys
from types import MethodType
from xml.etree import ElementTree

from qeschema.utils import set_logger, etree_iter_path, to_fortran, BiunivocalMap


class TestHelperFunctions(unittest.TestCase):

    def test_set_logger(self):
        logger = logging.getLogger('qeschema')
        loglevel = logger.getEffectiveLevel()
        handlers = logger.handlers

        set_logger(4)
        self.assertEqual(logger.getEffectiveLevel(), logging.DEBUG)
        self.assertIsInstance(logger.handlers[0], logging.StreamHandler)
        self.assertNotIsInstance(logger.handlers[0], logging.FileHandler)
        logger.handlers = []

        set_logger(3)
        self.assertEqual(logger.getEffectiveLevel(), logging.INFO)
        self.assertIsInstance(logger.handlers[0], logging.StreamHandler)
        self.assertNotIsInstance(logger.handlers[0], logging.FileHandler)
        logger.handlers = []

        set_logger(2)
        self.assertEqual(logger.getEffectiveLevel(), logging.WARNING)
        self.assertIsInstance(logger.handlers[0], logging.StreamHandler)
        self.assertNotIsInstance(logger.handlers[0], logging.FileHandler)
        self.assertEqual(len(logger.handlers), 1)

        set_logger(2)
        self.assertEqual(logger.getEffectiveLevel(), logging.WARNING)
        self.assertIsInstance(logger.handlers[0], logging.StreamHandler)
        self.assertNotIsInstance(logger.handlers[0], logging.FileHandler)
        self.assertEqual(len(logger.handlers), 1)

        logger.handlers = []

        f = tempfile.mkstemp()
        set_logger(1, logfile=f[1])
        self.assertEqual(logger.getEffectiveLevel(), logging.ERROR)
        self.assertIsInstance(logger.handlers[0], logging.FileHandler)
        logger.handlers[0].close()
        logger.setLevel(loglevel)
        logger.handlers = handlers

    def test_etree_iterpath(self):
        root = ElementTree.XML('<A><B1><C1/></B1><B2><C2/><C3/></B2></A>')

        values = iter(etree_iter_path(root))
        self.assertEqual(next(values), (root, '.'))
        self.assertEqual(next(values), (root[0], './B1'))
        self.assertEqual(next(values), (root[0][0], './B1/C1'))
        self.assertEqual(next(values), (root[1], './B2'))
        self.assertEqual(next(values), (root[1][0], './B2/C2'))
        self.assertEqual(next(values), (root[1][1], './B2/C3'))
        self.assertRaises(StopIteration, next, values)

        values = iter(etree_iter_path(root, tag='C2'))
        self.assertEqual(next(values), (root[1][0], './B2/C2'))
        self.assertRaises(StopIteration, next, values)

        values = iter(etree_iter_path(root, tag="*"))
        self.assertEqual(next(values), (root, '.'))
        self.assertEqual(next(values), (root[0], './B1'))
        self.assertEqual(next(values), (root[0][0], './B1/C1'))
        self.assertEqual(next(values), (root[1], './B2'))
        self.assertEqual(next(values), (root[1][0], './B2/C2'))
        self.assertEqual(next(values), (root[1][1], './B2/C3'))
        self.assertRaises(StopIteration, next, values)

        values = iter(etree_iter_path(root, path='/A'))
        self.assertEqual(next(values), (root, '/A'))
        self.assertEqual(next(values), (root[0], '/A/B1'))
        self.assertEqual(next(values), (root[0][0], '/A/B1/C1'))
        self.assertEqual(next(values), (root[1], '/A/B2'))
        self.assertEqual(next(values), (root[1][0], '/A/B2/C2'))
        self.assertEqual(next(values), (root[1][1], '/A/B2/C3'))
        self.assertRaises(StopIteration, next, values)

    def test_to_fortran(self):
        self.assertEqual(to_fortran(True), '.true.')
        self.assertEqual(to_fortran(False), '.false.')
        self.assertEqual(to_fortran(' a string '), repr('a string'))
        self.assertEqual(to_fortran(b' a byte string '), repr('a byte string'))
        self.assertEqual(to_fortran(10), '10')
        self.assertEqual(to_fortran(999.1), '999.1')


class TestBiunivocalMap(unittest.TestCase):

    def test_initialization(self):
        bimap = BiunivocalMap()
        self.assertFalse(bimap)
        self.assertFalse(getattr(bimap, '_BiunivocalMap__map'))
        self.assertFalse(getattr(bimap, '_BiunivocalMap__inverse'))

        bimap = BiunivocalMap([('a', 1), ('b', 2)])
        self.assertTrue(bimap)
        self.assertEqual(bimap, {'a': 1, 'b': 2})
        self.assertEqual(getattr(bimap, '_BiunivocalMap__map'), {'a': 1, 'b': 2})
        self.assertEqual(getattr(bimap, '_BiunivocalMap__inverse'), {1: 'a', 2: 'b'})

        bimap = BiunivocalMap({'a': 1, 'b': 2})
        self.assertTrue(bimap)
        self.assertEqual(bimap, {'a': 1, 'b': 2})
        self.assertEqual(getattr(bimap, '_BiunivocalMap__map'), {'a': 1, 'b': 2})
        self.assertEqual(getattr(bimap, '_BiunivocalMap__inverse'), {1: 'a', 2: 'b'})

    def test_setitem(self):
        bimap = BiunivocalMap({'a': 1, 'b': 2})
        bimap['c'] = 3
        bimap['d'] = 4

        result = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        self.assertEqual(bimap, result)
        self.assertEqual(getattr(bimap, '_BiunivocalMap__map'), result)
        self.assertEqual(getattr(bimap, '_BiunivocalMap__inverse'),
                         {1: 'a', 2: 'b', 3: 'c', 4: 'd'})

    def test_getitem(self):
        bimap = BiunivocalMap({'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(bimap['a'], 1)
        self.assertEqual(bimap['b'], 2)
        self.assertEqual(bimap['c'], 3)
        with self.assertRaises(KeyError):
            _ = bimap['d']

        bimap.__missing__ = MethodType(lambda x, y: 0, bimap)
        self.assertEqual(bimap['d'], 0)

    def test_delitem(self):
        bimap = BiunivocalMap({'a': 1, 'b': 2, 'c': 3, 'd': 4})

        del bimap['c']
        bimap.pop('d')
        self.assertEqual(bimap, {'a': 1, 'b': 2})
        self.assertEqual(getattr(bimap, '_BiunivocalMap__map'), {'a': 1, 'b': 2})
        self.assertEqual(getattr(bimap, '_BiunivocalMap__inverse'), {1: 'a', 2: 'b'})

    def test_duplicates(self):
        with self.assertRaises(ValueError):
            BiunivocalMap([('a', 1), ('b', 1)])

        bimap = BiunivocalMap({'a': 1, 'b': 2})
        with self.assertRaises(ValueError):
            bimap['c'] = 1

        bimap['a'] = 1
        with self.assertRaises(ValueError):
            bimap['a'] = 2

        bimap['b'] = 2.0
        with self.assertRaises(ValueError):
            bimap['a'] = 2

        bimap['b'] = 2.1
        bimap['a'] = 2

    def test_iter(self):
        bimap = BiunivocalMap({'a': 1, 'b': 2, 'c': 3, 'd': 4})

        if sys.version_info >= (3, 6):
            self.assertListEqual(list(x for x in bimap), ['a', 'b', 'c', 'd'])
        else:
            self.assertEqual(set(x for x in bimap), {'a', 'd', 'b', 'c'})

    def test_len(self):
        self.assertEqual(len(BiunivocalMap()), 0)
        self.assertEqual(len(BiunivocalMap({'a': 1, 'b': 2, 'c': 3, 'd': 4})), 4)

    def test_contains(self):
        bimap = BiunivocalMap({'a': 1, 'b': 2, 'c': 3, 'd': 4})
        self.assertTrue('a' in bimap)
        self.assertTrue('c' in bimap)
        self.assertTrue('d' in bimap)
        self.assertFalse('e' in bimap)
        self.assertFalse(4 in bimap)

    def test_repr(self):
        bimap = BiunivocalMap()
        self.assertEqual(repr(bimap), "BiunivocalMap()")

        bimap = BiunivocalMap({'a': 1, 'b': 2, 'c': 3, 'd': 4})
        if sys.version_info >= (3, 6):
            self.assertEqual(repr(bimap), "BiunivocalMap({'a': 1, 'b': 2, 'c': 3, 'd': 4})")
        else:
            s = repr(bimap)
            self.assertTrue(s.startswith("BiunivocalMap({"))
            self.assertTrue(s.endswith('})'))
            self.assertListEqual(sorted(s[15:-2].split(', ')),
                                 ["'a': 1", "'b': 2", "'c': 3", "'d': 4"])

    def test_copy(self):
        bimap = BiunivocalMap({'a': 1, 'b': 2, 'c': 3})
        bimap2 = bimap.copy()
        self.assertEqual(bimap, bimap2)

        self.assertEqual(bimap, {'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(getattr(bimap, '_BiunivocalMap__map'), {'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(getattr(bimap, '_BiunivocalMap__inverse'), {1: 'a', 2: 'b', 3: 'c'})

    def test_inverse(self):
        bimap = BiunivocalMap({'a': 1, 'b': 2, 'c': 3})
        self.assertEqual(bimap.inverse(), {1: 'a', 2: 'b', 3: 'c'})

    def test_getkey(self):
        bimap = BiunivocalMap({'a': 1, 'b': 2, 'c': 3})

        self.assertEqual(bimap.getkey(1), 'a')
        self.assertEqual(bimap.getkey(2), 'b')
        self.assertEqual(bimap.getkey(3), 'c')

        self.assertIsNone(bimap.getkey(4))
        self.assertEqual(bimap.getkey(4, 'unknown'), 'unknown')


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python3
#
# Copyright (c), 2015-2019, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
# Authors: Davide Brunato
#
import unittest
import logging

from qeschema import XmlDocumentError
from qeschema.options import get_specie_related_values, get_starting_magnetization, \
    get_system_nspin, set_ibrav_to_zero, get_system_eamp, get_electrons_efield, \
    get_system_edir, get_electric_potential_related, get_control_gdir, \
    get_cell_dofree, neb_set_system_nat, ha2ry, set_one_amass_line, \
    set_lda_plus_u_flag, set_boolean_flag, set_what_td_calculation

logger = logging.getLogger('qeschema')


class TestConversionFunctions(unittest.TestCase):

    def test_get_specie_related_values(self):
        kwargs = {
            'atomic_species': {
                '@ntyp': 3,
                'species': [
                    {'@name': 'O1', 'mass': 1.0, 'pseudo_file': 'O.pz-rrkjus.UPF',
                     'starting_magnetization': 0.0},
                    {'@name': 'Fe1', 'mass': 1.0, 'pseudo_file': 'Fe.pz-nd-rrkjus.UPF',
                     'starting_magnetization': 0.5},
                    {'@name': 'Fe2', 'mass': 1.0, 'pseudo_file': 'Fe.pz-nd-rrkjus.UPF',
                     'starting_magnetization': -0.5}]},
            'Hubbard_U': [{'@specie': 'O1', '@label': 'no Hubbard', '$': 0.0},
                          {'@specie': 'Fe1', '@label': '3d', '$': 0.3160442},
                          {'@specie': 'Fe2', '@label': '3d', '$': 0.3160442}],
            '_related_tag': 'Hubbard_U'
        }
        result = get_specie_related_values('Hubbard_U', **kwargs)
        self.assertListEqual(result, [' Hubbard_U(2)=0.3160442', ' Hubbard_U(3)=0.3160442'])

        kwargs['Hubbard_U'][1]['@spin'] = True
        result = get_specie_related_values('Hubbard_U', **kwargs)
        self.assertListEqual(result, [' Hubbard_U(2)=0.3160442', ' Hubbard_U(3)=0.3160442'])

        kwargs['atomic_species']['species'][1]['@name'] = 'unknown'
        with self.assertRaises(XmlDocumentError) as ctx:
            get_specie_related_values('Hubbard_U', **kwargs)
        self.assertEqual(str(ctx.exception), "Unknown specie 'Fe1' in tag 'Hubbard_U'")

        del kwargs['atomic_species']['species'][1]['@name']
        with self.assertRaises(KeyError):
            get_specie_related_values('Hubbard_U', **kwargs)

        del kwargs['atomic_species']
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_specie_related_values('Hubbard_U', **kwargs)

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required argument "
                                          "'atomic_species' when building parameter 'Hubbard_U'"])

        kwargs = {
            'atomic_species': {
                '@ntyp': 3,
                'species': [
                    {'@name': 'O1', 'mass': 1.0, 'pseudo_file': 'O.pz-rrkjus.UPF',
                     'starting_magnetization': 0.0},
                    {'@name': 'Fe1', 'mass': 1.0, 'pseudo_file': 'Fe.pz-nd-rrkjus.UPF',
                     'starting_magnetization': 0.5},
                    {'@name': 'Fe2', 'mass': 1.0, 'pseudo_file': 'Fe.pz-nd-rrkjus.UPF',
                     'starting_magnetization': -0.5}]
            },
            'Hubbard_J': [{'@specie': 'O1', '@label': 'no Hubbard', '$': [1.0, 0.0, 0.0]},
                          {'@specie': 'Fe1', '@label': '3d', '$': [0.0, 1.0, 0.0]},
                          {'@specie': 'Fe2', '@label': '3d', '$': [0.0, 0.0, 1.0]}],
            '_related_tag': 'Hubbard_J'
        }

        result = get_specie_related_values('Hubbard_J', **kwargs)
        self.assertListEqual(result, [' Hubbard_J(2,2)=1.0', ' Hubbard_J(3,3)=1.0'])

    def test_get_starting_magnetization(self):
        kwargs = {
            'atomic_species': {
                'species': [{'@name': 'Al', 'mass': 26.981538, 'pseudo_file': 'Al.pbe-n-van.UPF'},
                            {'@name': 'H', 'mass': 1.00794, 'pseudo_file': 'H.pbe-van_ak.UPF'}]
            },
            '_related_tag': 'atomic_species'
        }
        result = get_starting_magnetization('starting_magnetization', **kwargs)
        self.assertListEqual(result, [' starting_magnetization(1)=0.0', ' starting_magnetization(2)=0.0'])

        del kwargs['atomic_species']
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_starting_magnetization('starting_magnetization', **kwargs)

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments when building "
                                          "parameter 'starting_magnetization'! 'atomic_species'"])

    def test_get_system_nspin(self):
        kwargs = {'lsda': True, '_related_tag': 'noncolin', 'noncolin': False}

        result = get_system_nspin('nspin', **kwargs)
        self.assertListEqual(result, [' nspin=2'])

        del kwargs['lsda']
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_system_nspin('nspin', **kwargs)

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments "
                                          "when building parameter 'nspin'! 'lsda'"])

    def test_set_ibrav_to_zero(self):
        self.assertListEqual(set_ibrav_to_zero('name'), [' ibrav=0'])

    def test_get_system_eamp(self):
        kwargs = {'electric_potential': 'Berry_Phase', '_related_tag': 'electric_potential'}

        result = get_system_eamp('eamp', **kwargs)
        self.assertListEqual(result, [])

        kwargs['electric_potential'] = 'sawtooth_potential'
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_system_eamp('eamp', **kwargs)

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments when "
                                          "building parameter 'eamp'! 'electric_field_amplitude'"])

        kwargs['electric_field_amplitude'] = '0.0000000E+00'
        result = get_system_eamp('eamp', **kwargs)
        self.assertListEqual(result, [' eamp=0.0000000E+00'])

        kwargs['electric_potential'] = 'unknown'
        result = get_system_eamp('eamp', **kwargs)
        self.assertListEqual(result, [])

    def test_get_electrons_efield(self):
        kwargs = {'electric_potential': 'Berry_Phase', '_related_tag': 'electric_potential'}

        result = get_electrons_efield('efield', **kwargs)
        self.assertListEqual(result, [])

        kwargs['electric_potential'] = 'homogenous_field'
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_electrons_efield('efield', **kwargs)

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments when "
                                          "building parameter 'efield'! 'electric_field_amplitude'"])

        kwargs['electric_field_amplitude'] = '0.0000000E+00'
        result = get_electrons_efield('efield', **kwargs)
        self.assertListEqual(result, [' efield=0.0000000E+00'])

        kwargs['electric_potential'] = 'unknown'
        result = get_electrons_efield('efield', **kwargs)
        self.assertListEqual(result, [])

    def test_get_system_edir(self):
        kwargs = {'electric_potential': 'Berry_Phase',
                  '_related_tag': 'electric_field_direction',
                  'electric_field_direction': 3}

        result = get_system_edir('edir', **kwargs)
        self.assertListEqual(result, [])

        kwargs['electric_potential'] = 'sawtooth_potential'
        result = get_system_edir('edir', **kwargs)
        self.assertListEqual(result, [' edir=3'])

        del kwargs['electric_potential']
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_system_edir('edir', **kwargs)

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments when "
                                          "building parameter 'edir'! 'electric_potential'"])

    def test_get_electric_potential_related(self):
        result = get_electric_potential_related('tefield', electric_potential='Berry_Phase')
        self.assertListEqual(result, [' tefield=.false.'])

        result = get_electric_potential_related('lelfield', electric_potential='Berry_Phase')
        self.assertListEqual(result, [' lelfield=.false.'])

        result = get_electric_potential_related('lberry', electric_potential='Berry_Phase')
        self.assertListEqual(result, [' lberry=.true.'])

        result = get_electric_potential_related('unknown', electric_potential='Berry_Phase')
        self.assertListEqual(result, [])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_electric_potential_related('lberry')

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments when "
                                          "building parameter 'lberry'! 'electric_potential'"])

    def test_get_control_gdir(self):
        kwargs = {'electric_potential': 'Berry_Phase',  'electric_field_direction': 3}

        result = get_control_gdir('gdir', **kwargs)
        self.assertListEqual(result, [' gdir=3'])

        kwargs['electric_potential'] = 'other'
        result = get_control_gdir('gdir', **kwargs)
        self.assertListEqual(result, [])

        del kwargs['electric_potential']
        with self.assertLogs(logger, level='ERROR') as context:
            result = get_control_gdir('gdir', **kwargs)

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ["ERROR:qeschema:Missing required arguments when "
                                          "building parameter 'gdir'! 'electric_potential'"])

    def test_get_cell_dofree(self):
        result = get_cell_dofree('cell_dofree')
        self.assertListEqual(result, [ " cell_dofree = 'all'"])

        result = get_cell_dofree('cell_dofree', fix_volume=True)
        self.assertListEqual(result, [" cell_dofree = 'shape'"])

        result = get_cell_dofree('cell_dofree', fix_area=True)
        self.assertListEqual(result, [" cell_dofree = '2Dshape'"])

        result = get_cell_dofree('cell_dofree', isotropic=True)
        self.assertListEqual(result, [" cell_dofree = 'volume'"])

        result = get_cell_dofree('cell_dofree',cell_do_free="ibrav+volume")
        self.assertListEqual(result, [" cell_dofree = 'ibrav+volume'"])

        with self.assertLogs(logger, level='ERROR') as context:
            result = get_cell_dofree('cell_dofree', fix_volume=True, fix_area=True)

        self.assertListEqual(result, [" cell_dofree = 'all'"])
        err = ('ERROR:qeschema:only one of fix_volume, fix_area, fix_xy, '
               'isotropic, cell_do_free can be true')
        self.assertEqual(context.output, [err])

        with self.assertLogs(logger, level='ERROR' ) as context:
          result = get_cell_dofree('cell_dofree', fix_volume=True, cell_do_free="volume")
        err = ('ERROR:qeschema:only one of fix_volume, fix_area, fix_xy, '
               'isotropic, cell_do_free can be true')
        self.assertEqual(context.output, [err])

    def test_neb_set_system_nat(self):
        with self.assertLogs(logger, level='ERROR') as context:
            result = neb_set_system_nat('nat')

        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema:No atomic_structure element found !!!'])

        result = neb_set_system_nat('nat', atomic_structure=[{'@nat': 5}])
        self.assertListEqual(result, [' nat = 5'])

        with self.assertLogs(logger, level='ERROR') as context:
            result = neb_set_system_nat('nat', atomic_structure=[{}])
        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema:error reading nat value from atomic_structure !!!'])

        with self.assertLogs(logger, level='ERROR') as context:
            result = neb_set_system_nat('nat', atomic_structure=[])
        self.assertListEqual(result, [])
        self.assertEqual(context.output, ['ERROR:qeschema:error reading nat value from atomic_structure !!!'])

    def test_ha2ry(self):
        result = ha2ry('alpha', _related_tag='value', value=10)
        self.assertListEqual(result, [' alpha =  20.00000000'])

    def test_set_one_amass_line(self):
        result = set_one_amass_line('amass', amass={'@atom': 1, '$': 26.98})
        self.assertListEqual(result, [' amass(1)= 26.980'])

        result = set_one_amass_line('amass', amass=[{'@atom': 1, '$': 26.98}, {'@atom': 2, '$': 74.92}])
        self.assertListEqual(result, [' amass(1)= 26.980', ' amass(2)= 74.920'])

    def test_set_lda_plus_u_flag(self):
        kwargs = {
            '_related_tag': 'Hubbard_U',
            'Hubbard_U': [{'@specie': 'O1', '@label': 'no Hubbard', '$': 0.0},
                          {'@specie': 'Fe1', '$': 0.3160442},
                          {'@specie': 'Fe2', '$': 0.3160442}],
        }

        result = set_lda_plus_u_flag('lda_plus_u', **kwargs)
        self.assertListEqual(result, [' lda_plus_u = .true.'])

        result = set_lda_plus_u_flag('lda_plus', _related_tag='Hubbard_U', Hubbard_U=[])
        self.assertListEqual(result, [])

    def test_set_boolean_flag(self):
        result = set_boolean_flag('restart', restart=False, _related_tag='restart')
        self.assertListEqual(result, [' restart = .false.'])

        result = set_boolean_flag('pseudo_hermitian', pseudo_hermitian=True, _related_tag='pseudo_hermitian')
        self.assertListEqual(result, [' pseudo_hermitian = .true.'])

    def test_set_what_td_calculation(self):
        result = set_what_td_calculation('what', whatTD='davidson')
        self.assertListEqual(result, ['davidson'])

        with self.assertRaises(AssertionError):
            set_what_td_calculation(True, whatTD='davidson')

        with self.assertRaises(KeyError):
            set_what_td_calculation('what')


if __name__ == '__main__':
    unittest.main()



#
# Copyright (c), 2015-2022, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato
#
from .documents import XmlDocument, QeDocument, PwDocument, PhononDocument, \
    NebDocument, TdDocument, TdSpectrumDocument, XSpectraDocument, EPWDocument
from .converters import RawInputConverter, PwInputConverter, \
    PhononInputConverter, NebInputConverter, TdInputConverter, \
    TdSpectrumInputConverter, XSpectraInputConverter, EPWInputConverter
from .exceptions import QESchemaError, XmlDocumentError
from .utils import set_logger

__version__ = '1.5.1'

__all__ = [
    'XmlDocument', 'QeDocument', 'PwDocument', 'PhononDocument', 'NebDocument',
    'TdDocument', 'TdSpectrumDocument', 'EPWDocument', 'RawInputConverter',
    'PwInputConverter', 'PhononInputConverter', 'TdInputConverter',
    'TdSpectrumInputConverter', 'NebInputConverter', 'QESchemaError',
    'XmlDocumentError', 'set_logger', 'hdf5', 'XSpectraDocument',
    'XSpectraInputConverter', 'EPWInputConverter'
]



#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato, Giovanni Borghi
#
"""
Data format converters for Quantum Espresso.
"""
import copy
import logging
import re
import os.path
from collections.abc import Container

from .utils import to_fortran, BiunivocalMap
from . import cards, options

logger = logging.getLogger('qeschema')

FCP_NAMELIST = 'FCP'
PH_NAT_TODO_CARD = 'ph_nat_todo_card'
OPTIONAL_CARDS = {PH_NAT_TODO_CARD, "CONSTRAINTS", "HUBBARD", "SOLVENTS"}


def conversion_maps_builder(template_map):
    """
    Build invariant and variant conversion maps from a template. The template
    is a multilevel dictionary that reproduce the structure of the input data.
    An entry maybe:

      1) a string with the list and the option names (eg: "CONTROL[calculation]").
         In this case the entry is mapped to this value, without modifications.

      2) a tuple with three items:
         - list and option names (eg: 'SYSTEM[Hubbard_U]');
         - the function that has to be used to calculate the namelist option
           value from XML value;
         - the function that has to be used to calculate to calculate the XML
           value data from namelist entry.

      3) a list of tuples when the XML entry has to be used to calculate other
         namelist or card's entries. Each tuple has the meaning of the previous
         case. The second and third elements may be None or empty if the template
         has another tuple with the same list and option names that indicate the
         transformation functions associated.

    :param template_map: Template dictionary
    :return: A couple of dictionaries, the first for invariant
    parameter mappings, se second for multi-dependent or type
    transformations.
    """
    def _check_variant(*args):
        if not args:
            raise ValueError("Empty sequence not admitted for a conversion map!")
        elif not isinstance(args[0], str):
            raise TypeError("The first element must be a string! '{0}'".format(args))
        elif len(args) == 1:
            return args[0], None, None
        elif len(args) == 2 and (callable(args[1]) or args[1] is None):
            return args[0], args[1], None
        elif len(args) == 3 and (callable(args[1]) or args[1] is None) \
                and (callable(args[2]) or args[2] is None):
            return args
        else:
            raise TypeError("Need a callable or None for 2nd and 3rd items! '{0}'".format(*args))

    def _build_maps(dict_element, path):
        for key, item in dict_element.items():
            path_key = '/'.join((path, key))
            if isinstance(item, dict):
                _build_maps(item, path_key)

            elif isinstance(item, str):
                invariant_map[path_key] = item
                logger.debug("Added one-to-one association: %r=%r", path_key, item)

            elif isinstance(item, (tuple, list)):
                try:
                    variant_map[path_key] = _check_variant(*item)
                    logger.debug("Added single-variant mapping: %r=%r",
                                 path_key, variant_map[path_key])
                except TypeError:
                    variants = []
                    for variant in item:
                        if isinstance(variant, str):
                            invariant_map[path_key] = variant
                        elif isinstance(variant, (tuple, list)):
                            variants.append(_check_variant(*variant))
                        else:
                            raise TypeError(f"Expect a tuple, list or string! {variant!r}")

                    variant_map[path_key] = tuple(variants)
                    logger.debug("Added multi-variant mapping: %r=%r",
                                 path_key, variant_map[path_key])

    invariant_map = BiunivocalMap()
    variant_map = dict()
    if not isinstance(template_map, dict):
        raise TypeError("A dictionary is needed!")
    _build_maps(template_map, '.')

    # Check inconsistencies between maps
    for items in variant_map.values():
        for value in items:
            logger.debug("Check value: %r", value)
            if isinstance(value, str) and value in invariant_map.inverse():
                raise ValueError("A variant is also in invariant map! "
                                 "'%s': '%s'" % (invariant_map.getkey(value), value))
            elif isinstance(value, tuple) and value[0] in invariant_map.inverse():
                raise ValueError("A variant is also in invariant map! "
                                 "'%s': '%s'" % (invariant_map.getkey(value[0]), value[0]))

    return invariant_map, variant_map


class RawInputConverter(Container):
    """
    A Fortran's namelist builder.
    """
    target_pattern = re.compile(r'(\w+)(?:\[((?:\w+)(?:%\w+)*)]|)')
    """RE pattern to extract Fortran input's namelist/card and name of a parameter"""

    def __init__(self, invariant_map, variant_map, input_namelists=None, input_cards=None):
        self.invariant_map = invariant_map
        """Map of parameters that are matched with one-to-one relation
        and which value is invariant from XML to QE Fortran input format."""

        self.variant_map = variant_map
        """Map of parameters that require a type conversion or are multi
        related with other inputs."""

        self.input_namelists = tuple(input_namelists) or (
            key for key in invariant_map.keys() + variant_map.keys()
        )
        """Sequence of input namelists. Use keys of the maps if not provided."""
        if input_cards:
            self.input_cards = tuple(input_cards)
        else:
            self.input_cards = ()
        """Sequence of input special cards, assuming an empty list if not provided."""

        self._input = dict(
            [(section, {}) for section in self.input_namelists + self.input_cards]
        )

    def __contains__(self, path):
        return path in self.invariant_map or path in self.variant_map

    def set_path(self, path, tag, node_dict):
        """
        Insert values for a path.

        :param path:
        :param tag:
        :param node_dict:
        :return:
        """
        if len(node_dict) != 1:
            raise ValueError("The node_dict argument must contains exactly "
                             "one element! {0}".format(node_dict))
        logger.debug("Set input with path %r and node dict %r", path, node_dict)
        _path, _, keyword = path.rpartition('/')
        value = node_dict[tag]
        if isinstance(value, dict) and keyword != tag:
            try:
                value = value[keyword]
            except KeyError:
                if keyword == '$':
                    value = node_dict[tag]
                else:
                    raise ValueError(
                        "Keyword '{0}' not found in node_dict \"{1}\"!".format(keyword, node_dict)
                    )

        if value is None:
            logger.debug("Skip element %r: None value!", path)
            return

        # Set the target parameter if the path is in invariant_map dictionary
        if not isinstance(value, dict) and path in self.invariant_map:
            self.set_parameter(path, value)

        # Add argument to variant transformations associated with the path
        if path in self.variant_map:
            self.add_kwarg(path, tag, node_dict)

    def set_parameter(self, path, value):
        target = self.invariant_map[path]
        match = self.target_pattern.match(target)
        if match is not None:
            namelist, name = match.groups()
        else:
            namelist = name = None
        if name is None:
            raise ValueError("Wrong value {!r} for invariant parameter {!r}".format(target, path))

        self._input[namelist][name] = to_fortran(value)
        logger.debug("Set %s[%s]=%s", namelist, name, self._input[namelist][name])

    def add_kwarg(self, path, tag, node_dict):
        if isinstance(self.variant_map[path][0], str):
            target_items = list([self.variant_map[path]])[:2]
        else:
            target_items = self.variant_map[path]
        for target, _get_qe_input, _ in target_items:
            logger.debug("Add argument to %r", target)
            logger.debug("Argument's conversion function: %r", _get_qe_input)
            group, name = self.target_pattern.match(target).groups()
            if name is not None:
                try:
                    name_dict = self._input[group][name]
                except KeyError:
                    self._input[group][name] = node_dict.copy()
                else:
                    try:
                        name_dict[tag].append(node_dict[tag])
                    except AttributeError:
                        name_dict[tag] = [self._input[group][name][tag], node_dict[tag]]
                    except KeyError:
                        name_dict.update(node_dict.copy())
                if _get_qe_input is not None:
                    self._input[group][name].update({
                        '_get_qe_input': _get_qe_input,
                        '_related_tag': tag
                    })
            else:
                try:
                    self._input[group][tag].append(node_dict[tag])
                except AttributeError:
                    self._input[group][tag] = [self._input[group][tag], node_dict[tag]]
                except KeyError:
                    self._input[group].update(node_dict.copy())
                if _get_qe_input is not None:
                    self._input[group].update({'_get_qe_input': _get_qe_input})

    def get_qe_input(self):
        if all([not section for section in self._input.values()]):
            logger.error("Empty input!")
        _input = self._input
        lines = []
        for namelist in self.input_namelists:

            if namelist == FCP_NAMELIST and not len(_input[namelist]):
                # Skip empty &FCP/ section
                continue

            lines.append('&%s' % namelist)
            for name, value in sorted(_input[namelist].items(), key=lambda x: x[0].lower()):
                logger.debug("Add input for parameter %s[%r] with value %r",
                             namelist, name, value)

                if isinstance(value, dict):
                    # Variant conversion: apply to_fortran_input function with saved arguments
                    try:
                        to_fortran_input = value['_get_qe_input']
                    except KeyError:
                        logger.debug('No conversion function for parameter %s[%r], skip ... ',
                                     namelist, name)
                        continue

                    if callable(to_fortran_input):
                        lines.extend(to_fortran_input(name, **value))
                    else:
                        logger.error('Parameter %s[%r] conversion function is not callable!',
                                     namelist, name)
                else:
                    # Simple invariant conversion
                    lines.append(' {0}={1}'.format(name, value))
            lines.append('/')

        for card in self.input_cards:
            logger.debug("Add card %r", card)
            card_args = _input[card]
            logger.debug("Card arguments: %r", card_args)

            if card not in OPTIONAL_CARDS and \
                    ('_get_qe_input' not in card_args or
                     not callable(card_args['_get_qe_input'])):
                logger.error("Missing conversion function for card %r", card)

            _get_qe_input = card_args.get('_get_qe_input', None)

            if callable(_get_qe_input):
                lines.extend(_get_qe_input(card, **card_args))
            elif card not in OPTIONAL_CARDS:
                logger.error('Card conversion function not found!')

        return '\n'.join(lines)

    def clear_input(self):
        del self._input
        self._input = dict(
            [(section, {}) for section in self.input_namelists + self.input_cards]
        )


class PwInputConverter(RawInputConverter):
    """
    Builds a Fortran's namelist input for PWscf.
    """
    PW_TEMPLATE_MAP = {
        'control_variables': {
            'title': "CONTROL[title]",
            'calculation': "CONTROL[calculation]",
            'restart_mode': "CONTROL[restart_mode]",
            'prefix': "CONTROL[prefix]",
            'pseudo_dir': "CONTROL[pseudo_dir]",
            'outdir': "CONTROL[outdir]",
            'stress': "CONTROL[tstress]",
            'forces': "CONTROL[tprnfor]",
            'wf_collect': "CONTROL[wf_collect]",
            'disk_io': "CONTROL[disk_io]",
            'max_seconds': "CONTROL[max_seconds]",
            'etot_conv_thr': "CONTROL[etot_conv_thr]",
            'forc_conv_thr': "CONTROL[forc_conv_thr]",
            'press_conv_thr': "CELL[press_conv_thr]",
            'verbosity': "CONTROL[verbosity]",
            'print_every': "CONTROL[iprint]",
            'nstep': "CONTROL[nstep]",
        },
        # Card ATOMIC species with attributes
        'atomic_species': {
            '@ntyp': 'SYSTEM[ntyp]',
            '$': [
                ("ATOMIC_SPECIES", cards.get_atomic_species_card, None),
                ('SYSTEM[Hubbard_U]',),
                ('SYSTEM[Hubbard_J0]',),
                ('SYSTEM[Hubbard_alpha]',),
                ('SYSTEM[Hubbard_beta]',),
                ('SYSTEM[Hubbard_J]',),
                ('SYSTEM[starting_ns_eigenvalue]',),
                ('SYSTEM[starting_magnetization]', options.get_starting_magnetization, None),
            ]
        },
        'atomic_structure': {
            '@nat': 'SYSTEM[nat]',
            '$': [
                ('SYSTEM[ibrav]', options.set_ibrav_to_zero, None),
                ("ATOMIC_POSITIONS", cards.get_atomic_positions_cell_card, None),
                ("CELL_PARAMETERS", cards.get_cell_parameters_card,  None)
            ],
            'atomic_positions': ('ATOMIC_FORCES', cards.get_atomic_forces_card, None),
            'crystal_positions': ('ATOMIC_FORCES', cards.get_atomic_forces_card, None),
            'wyckoff_positions': ('ATOMIC_FORCES', cards.get_atomic_forces_card, None)
        },
        'dft': {
            'functional': "SYSTEM[input_dft]",
            'hybrid': {
                'qpoint_grid': {
                    '@nqx1': 'SYSTEM[nqx1]',
                    '@nqx2': 'SYSTEM[nqx2]',
                    '@nqx3': 'SYSTEM[nqx3]'
                },
                'ecutfock': ('SYSTEM[ecutfock]', options.ha2ry, None),
                'exx_fraction': 'SYSTEM[exx_fraction]',
                'screening_parameter': 'SYSTEM[screening_parameter]',
                'exxdiv_treatment': 'SYSTEM[exxdiv_treatment]',
                'x_gamma_extrapolation': 'SYSTEM[x_gamma_extrapolation]',
                'ecutvcut': ('SYSTEM[ecutvcut]', options.ha2ry, None)
            },
            'dftU': {
                '@new_format': [('HUBBARD', cards.get_hubbard_card, None),
                                ('SYSTEM[lda_plus_u_kind]', options.set_lda_plus_u_kind, None),
                                ('SYSTEM[lda_plus_u]', options.set_lda_plus_u_flag, None),
                                ('SYSTEM[Hubbard_U]', options.get_specie_related_values, None),
                                ('SYSTEM[Hubbard_J]', options.get_specie_related_values, None),
                                ('SYSTEM[Hubbard_J0]', options.get_specie_related_values, None),
                                ('SYSTEM[Hubbard_alpha]', options.get_specie_related_values, None),
                                ('SYSTEM[Hubbard_beta]', options.get_specie_related_values, None),
                                ('SYSTEM[U_projection_type]', options.set_u_projection_type, None)],
                'lda_plus_u_kind': ('SYSTEM[lda_plus_u_kind]', options.set_lda_plus_u_kind, None),
                'Hubbard_U': {
                    '$': [('SYSTEM[Hubbard_U]', options.get_specie_related_values, None),
                          ('HUBBARD', cards.get_hubbard_card, None),
                          ('SYSTEM[lda_plus_u]', options.set_lda_plus_u_flag, None)]
                },
                'Hubbard_J0': {
                    '$': [('SYSTEM[Hubbard_J0]', options.get_specie_related_values, None),
                          ('HUBBARD', cards.get_hubbard_card, None)]
                },
                'Hubbard_alpha': {
                    '$': [('SYSTEM[Hubbard_alpha]', options.get_specie_related_values, None),
                          ('HUBBARD', cards.get_hubbard_card, None)]
                },
                'Hubbard_beta': {
                    '$': [('SYSTEM[Hubbard_beta]', options.get_specie_related_values, None),
                          ('HUBBARD', cards.get_hubbard_card, None)]
                },
                'Hubbard_J': {
                    '$': [('SYSTEM[Hubbard_J]', options.get_specie_related_values, None),
                          ('HUBBARD', cards.get_hubbard_card, None)]
                },
                'starting_ns': {
                    '$': ('SYSTEM[starting_ns_eigenvalue]', options.get_specie_related_values, None)
                },
                'U_projection_type': ('SYSTEM[U_projection_type]',
                                      options.set_u_projection_type, None)
            },
            'vdW': {
                'vdw_corr': 'SYSTEM[vdw_corr]',
                'london_s6': 'SYSTEM[london_s6]',
                'ts_vdw_econv_thr': 'SYSTEM[ts_vdw_econv_thr]',
                'ts_vdw_isolated': 'SYSTEM[ts_vdw_isolated]',
                'london_rcut': 'SYSTEM[london_rcut]',
                'xdm_a1': 'SYSTEM[xdm_a1]',
                'xdm_a2': 'SYSTEM[xdm_a2]',
                'dftd3_version': 'SYSTEM[dftd3_version]',
                'dftd3_threebody': 'SYSTEM[dftd3_threebody]',
                'london_c6': {
                    '$': ('SYSTEM[london_c6]', options.get_specie_related_values, None),
                }
            }
        },
        'spin': {
            'lsda': ("SYSTEM[nspin]", options.get_system_nspin, None),
            'noncolin': [
                ("SYSTEM[nspin]", options.get_system_nspin, None),
                "SYSTEM[noncolin]"
            ],
            'spinorbit': "SYSTEM[lspinorb]"
        },
        'bands': {
            'nbnd': "SYSTEM[nbnd]",
            'smearing': {
                '$': "SYSTEM[smearing]",
                '@degauss': "SYSTEM[degauss]"
            },
            'tot_charge': "SYSTEM[tot_charge]",
            'tot_magnetization': "SYSTEM[tot_magnetization]",
            'occupations': {
                '$': "SYSTEM[occupations]"
            }
        },
        'basis': {
            'gamma_only': ('K_POINTS', cards.get_k_points_card, None),
            'ecutwfc': "SYSTEM[ecutwfc]",
            'ecutrho': "SYSTEM[ecutrho]",
            'fft_grid': {
                'nr1': "SYSTEM[nr1]",
                'nr2': "SYSTEM[nr2]",
                'nr3': "SYSTEM[nr3]",
            },
            'fft_smooth': {
                'nr1': "SYSTEM[nr1s]",
                'nr2': "SYSTEM[nr2s]",
                'nr3': "SYSTEM[nr3s]",
            },
            'fft_box': {
                'nr1': "SYSTEM[nr1b]",
                'nr2': "SYSTEM[nr2b]",
                'nr3': "SYSTEM[nr3b]",
            },
            'spline_ps': "SYSTEM[spline_ps]"
        },
        'electron_control': {
            'diagonalization': "ELECTRONS[diagonalization]",
            'mixing_mode': "ELECTRONS[mixing_mode]",
            'mixing_beta': "ELECTRONS[mixing_beta]",
            'conv_thr': "ELECTRONS[conv_thr]",
            'mixing_ndim': "ELECTRONS[mixing_ndim]",
            'max_nstep': "ELECTRONS[electron_maxstep]",
            'real_space_q': "ELECTRONS[tqr]",
            'tq_smoothing': "ELECTRONS[tq_smoothing]",
            'tbeta_smoothing': "ELECTRONS[tbeta_smoothing]",
            'diago_thr_init': "ELECTRONS[diago_thr_init]",
            'diago_full_acc': "ELECTRONS[diago_full_acc]",
            'diago_cg_maxiter': "ELECTRONS[diago_cg_maxiter]"
        },
        'k_points_IBZ': ('K_POINTS', cards.get_k_points_card, None),
        'ion_control': {
            'ion_dynamics': "IONS[ion_dynamics]",
            'upscale': "IONS[upscale]",
            'remove_rigid_rot': "IONS[remove_rigid_rot]",
            'refold_pos': "IONS[refold_pos]",
            'bfgs': {
                'trust_radius_min': "IONS[trust_radius_min]",
                'trust_radius_max': "IONS[trust_radius_max]",
                'trust_radius_init': "IONS[trust_radius_ini]",
                'w1': "IONS[w_1]",
                'w2': "IONS[w_2]",
                'ndim': "IONS[bfgs_ndim]"
            },
            'md': {
                'pot_extrapolation': "IONS[pot_extrapolation]",
                'wfc_extrapolation': "IONS[wfc_extrapolation]",
                'ion_temperature': "IONS[ion_temperature]",
                'timestep': "CONTROL[dt]",
                'tempw': "IONS[tempw]",
                'tolp': "IONS[tolp]",
                'deltaT': "IONS[delta_t]",
                'nraise': "IONS[nraise]"
            }
        },
        'cell_control': {
            'cell_dynamics': "CELL[cell_dynamics]",
            'wmass': "CELL[wmass]",
            'cell_factor': "CELL[cell_factor]",
            'pressure': "CELL[press]",
            'free_cell': ("CELL_PARAMETERS", cards.get_cell_parameters_card, None),
            'fix_volume': ("CELL[cell_dofree]", options.get_cell_dofree, None),
            'fix_area': ("CELL[cell_dofree]", options.get_cell_dofree, None),
            'fix_xy': ("CELL[cell_dofree]", options.get_cell_dofree, None),
            'isotropic': ("CELL[cell_dofree]", options.get_cell_dofree, None),
            'cell_do_free': ("CELL[cell_dofree]", options.get_cell_dofree, None),
        },
        'symmetry_flags': {
            'nosym': "SYSTEM[nosym]",
            'nosym_evc': "SYSTEM[nosym_evc]",
            "noinv": "SYSTEM[noinv]",
            'no_t_rev': "SYSTEM[no_t_rev]",
            'force_symmorphic': "SYSTEM[force_symmorphic]",
            'use_all_frac': "SYSTEM[use_all_frac]"
        },
        'boundary_conditions': {
            'assume_isolated': "SYSTEM[assume_isolated]",
            'esm': {
                'bc': "SYSTEM[esm_bc]",
                'nfit': "SYSTEM[esm_nfit]",
                'w': "SYSTEM[esm_w]",
                'efield': "SYSTEM[esm_efield]"
            },
            'fcp_opt': "CONTROL[lfcp]",
            'fcp_mu': "FCP[fcp_mu]"
        },
        'ekin_functional': {
            'ecfixed': "SYSTEM[ecfixed]",
            'qcutz': "SYSTEM[qcutz]",
            'q2sigma': "SYSTEM[q2sigma]"
        },
        'external_atomic_forces': {
            '$': ('ATOMIC_FORCES', cards.get_atomic_forces_card, None)
        },
        'free_positions': {
            '$': [("ATOMIC_POSITIONS", cards.get_atomic_positions_cell_card, None),
                  ("CELL_PARAMETERS",)]},
        'electric_field': {
            'electric_potential': [
                ("CONTROL[tefield]", options.get_electric_potential_related),
                ("CONTROL[lelfield]", options.get_electric_potential_related),
                ("CONTROL[lberry]", options.get_electric_potential_related),
                ("SYSTEM[eamp]", options.get_system_eamp),
                ("ELECTRONS[efield]", options.get_electrons_efield),
                ("SYSTEM[edir]", options.get_system_edir),
                ("CONTROL[gdir]", options.get_control_gdir)
            ],
            'dipole_correction': "CONTROL[dipfield]",
            'electric_field_direction': [
                ("SYSTEM[edir]", options.get_system_edir,),
                ("CONTROL[gdir]", options.get_control_gdir),
            ],
            'potential_max_position': "SYSTEM[emaxpos]",
            'potential_decrease_width': "SYSTEM[eopreg]",
            'electric_field_amplitude': [
                ("SYSTEM[eamp]", options.get_system_eamp,),
                ("ELECTRONS[efield]", options.get_electrons_efield),
            ],
            'electric_field_vector': "ELECTRONS[efield_cart]",
            'nk_per_string': "CONTROL[nppstr]",
            'n_berry_cycles': "CONTROL[nberrycyc]",
        },
        'atomic_constraints': ("CONSTRAINTS", cards.get_atomic_constraints_card),  # Card
        'spin_constraints': {
            'spin_constraints': "SYSTEM[constrained_magnetization]",
            'lagrange_multiplier': "SYSTEM[lambda]",
            'target_magnetization': "SYSTEM[fixed_magnetization]",
        }
    }

    def __init__(self, **kwargs):
        super(PwInputConverter, self).__init__(
            *conversion_maps_builder(self.PW_TEMPLATE_MAP),
            input_namelists=('CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS', 'CELL',
                             FCP_NAMELIST),
            input_cards=('ATOMIC_SPECIES', 'ATOMIC_POSITIONS', 'K_POINTS',
                         'CELL_PARAMETERS', 'ATOMIC_FORCES', 'CONSTRAINTS', 'SOLVENTS', 'HUBBARD')
        )
        if 'xml_file' in kwargs:
            self._input['CONTROL']['input_xml_schema_file'] = "{!r}".format(
                os.path.basename(kwargs['xml_file'])
            )

    def clear_input(self):
        super(PwInputConverter, self).clear_input()


class PhononInputConverter(RawInputConverter):
    """
    Convert to/from Fortran input for Phonon.
    """
    PHONON_TEMPLATE_MAP = {
        'xq': {
            '$': ('qPointsSpecs', cards.get_qpoints_card, None),
        },
        'scf_ph': {
            'tr2_ph': "INPUTPH[tr2_ph]",
            'niter_ph': "INPUTPH[niter_ph]",
            'alpha_mix': "INPUTPH[alpha_mix]",
            'nmix_ph': "INPUTPH[nmix_ph]",
        },
        'files': {
            'prefix': "INPUTPH[prefix]",
            'outdir': "INPUTPH[outdir]",
            'fildyn': "INPUTPH[fildyn]",
            'fildrho': "INPUTPH[fildrho]",
            'fildvscf': "INPUTPH[fildvscf]",
            'lqdir': "INPUTPH[lqdir]",
        },
        'control_ph': {
            'ldisp': ["INPUTPH[ldisp]", ('qPointsSpecs', cards.get_qpoints_card, None)],
            'epsil': "INPUTPH[epsil]",
            'trans': "INPUTPH[trans]",
            'zeu': "INPUTPH[zeu]",
            'zue': "INPUTPH[zue]",
            'elop': "INPUTPH[elop]",
            'fpol': "INPUTPH[fpol]",
            'lraman': "INPUTPH[lraman]",
            'search_sym': "INPUTPH[search_sym]",
        },
        'control_job': {
            'recover': "INPUTPH[recover]",
            'max_seconds': "INPUTPH[max_seconds]",
        },
        'control_diel': {
            'lrpa': "INPUTPH[lrpa]",
            'lnoloc': "INPUTPH[lnoloc]",
        },
        'control_qplot': {
            'qplot': [('qPointsSpecs', cards.get_qpoints_card, None), "INPUTPH[qplot]"],
            'q2d': "INPUTPH[q2d]",
            'q_in_band_form': "INPUTPH[q_in_band_form]"
        },
        'miscellanea': {
            'amass': {'$': ("INPUTPH[amass]", options.set_one_amass_line, None)},
            'verbosity': "INPUTPH[verbosity]",
            'reduce_io': "INPUTPH[reduce_io]",
            'low_directory_check': "INPUTPH[low_directory_check]",
            'nogg': "INPUTPH[nogg]",
            'nscf_MPgrid': {
                'nk1': "INPUTPH[nk1]",
                'nk2': "INPUTPH[nk2]",
                'nk3': "INPUTPH[nk3]",
                'k1': "INPUTPH[k1]",
                'k2': "INPUTPH[k2]",
                'k3': "INPUTPH[k3]",
            }
        },
        'irr_repr': {
            'start_q': "INPUTPH[start_q]",
            'last_q': "INPUTPH[last_q]",
            'start_irr': "INPUTPH[start_irr]",
            'last_irr': "INPUTPH[last_irr]",
            'nat_todo':
            {
                '@natom': 'INPUTPH[nat_todo]',
                '$': [(PH_NAT_TODO_CARD, cards.get_nat_todo_card, None)]
            },
            'modenum': "INPUTPH[modenum]",
            'only_init': "INPUTPH[only_init]",
            'ldiag': "INPUTPH[ldiag]",
        },
        'electron_phonon_options': {
            'electron_phonon': "INPUTPH[electron_phonon]",
            'dvscf_star': {
                'open': "INPUTPH[dvscf_star%open]",
                'dir': "INPUTPH[dvscf_star%dir]",
                'ext': "INPUTPH[dvscf_star%ext]",
                'basis': "INPUTPH[dvscf_star%basis]",
                'pat': "INPUTPH[dvscf_star%pat]",
            },
            'drho_star': {
                'open': "INPUTPH[drho_star%open]",
                'dir': "INPUTPH[drho_star%dir]",
                'ext': "INPUTPH[drho_star%ext]",
                'basis': "INPUTPH[drho_star%basis]",
                'pat': "INPUTPH[drho_star%pat]",
            }
        },
        'lraman_options': {
            'eth_rps': "INPUTPH[eth_rps]",
            'eth_ns': "INPUTPH[eth_ns]",
            'dek': "INPUTPH[dek]",
        },
        'q_points': {
            'grid': {
                '@nq1': "INPUTPH[nq1]",
                '@nq2': "INPUTPH[nq2]",
                '@nq3': "INPUTPH[nq3]"
            },
            'q_points_list': ('qPointsSpecs', cards.get_qpoints_card, None),
            'nqs': ('qPointsSpecs', cards.get_qpoints_card, None)
        }
    }

    def __init__(self, **_kwargs):
        super(PhononInputConverter, self).__init__(
            *conversion_maps_builder(self.PHONON_TEMPLATE_MAP),
            input_namelists=('INPUTPH',),
            input_cards=('qPointsSpecs', PH_NAT_TODO_CARD)
        )


class NebInputConverter(RawInputConverter):
    """
    Convert to/from Fortran input for Phonon.
    """
    NEB_TEMPLATE_MAP = {
        'path': {
            'restartMode': "PATH[restart_mode]",
            'stringMethod': "PATH[string_method]",
            'pathNstep': "PATH[nstep_path]",
            'numOfImages': "PATH[num_of_images]",
            'optimizationScheme': "PATH[opt_scheme]",
            'optimizationStepLength': "PATH[ds]",
            'elasticConstMax': "PATH[k_max]",
            'elasticConstMin': "PATH[k_min]",
            'pathThreshold': "PATH[path_thr]",
            'endImagesOptimizationFlag': "PATH[first_last_opt]",
            'minimumImageFlag': "PATH[minimum_image]",
            'temperature': "PATH[temp_req]",
            'climbingImage': [
                "PATH[CI_scheme]",
                ("CLIMBING_IMAGES", cards.get_climbing_images, None)
            ],
            'useMassesFlag': "PATH[use_masses]",
            'useFreezingFlag': "PATH[use_freezing]",
            'constantBiasFlag': "PATH[lfcp]",
            'targetFermiEnergy': "PATH[fcp_mu]",
            'totChargeFirst': "PATH[fcp_tot_charge_first]",
            'totChargeLast': "PATH[fcp_tot_charge_last]",
            'climbingImageIndex': ("CLIMBING_IMAGES", cards.get_climbing_images, None)
        }
    }

    def __init__(self, **_kwargs):
        engine_template_map = copy.deepcopy(PwInputConverter.PW_TEMPLATE_MAP)
        engine_template_map['atomic_structure'] = {
            '@nat': ("SYSTEM[nat]", options.neb_set_system_nat, None),
            '$': [
                ('SYSTEM[ibrav]', options.set_ibrav_to_zero, None),
                ("CELL_PARAMETERS", cards.get_neb_cell_parameters_card, None),
                ("ATOMIC_POSITIONS", cards.get_neb_images_positions_card, None)
            ],
            'atomic_positions': ('ATOMIC_FORCES', cards.get_atomic_forces_card, None),
            'crystal_positions': ('ATOMIC_FORCES', cards.get_atomic_forces_card, None),
            'wyckoff_positions': ('ATOMIC_FORCES', cards.get_atomic_forces_card, None)
        }
        engine_template_map['free_positions'] = {
            '$': ("ATOMIC_POSITIONS", cards.get_neb_images_positions_card, None)
        }
        self.NEB_TEMPLATE_MAP.update({'engine': engine_template_map})
        super(NebInputConverter, self).__init__(
            *conversion_maps_builder(self.NEB_TEMPLATE_MAP),
            input_namelists=('PATH', 'CONTROL', 'SYSTEM', 'ELECTRONS', 'IONS', 'CELL'),
            input_cards=('CLIMBING_IMAGES', 'ATOMIC_SPECIES', 'ATOMIC_POSITIONS', 'K_POINTS',
                         'CELL_PARAMETERS', 'ATOMIC_FORCES', 'CONSTRAINTS')
        )

    def get_qe_input(self):
        """
        Overrides method in RawInputConverter because few lines
        in between the namelists are requested for the NEB input.

        :return: a string containing the text input for NEB calculations
        """
        qe_input = super(NebInputConverter, self).get_qe_input().split('\n')
        qe_input = ['BEGIN', 'BEGIN_PATH_INPUT'] + qe_input
        index = qe_input.index('&CONTROL')
        qe_input = qe_input[:index] + ['END_PATH_INPUT', 'BEGIN_ENGINE_INPUT'] + qe_input[index:]
        qe_input += ['END_ENGINE_INPUT', 'END']
        return '\n'.join(qe_input)


class TdInputConverter(RawInputConverter):
    """
    converts qes_lr schema to fortran input for turbo_lanczos, turbo_davidson and turbo_eels
    """
    TD_TEMPLATE_MAP = {
        'whatTD': ('cache[what]', options.set_what_td_calculation, None),
        'lr_input': {
            'restart': ('lr_input[restart]', options.set_boolean_flag, None),
            'restart_step': 'lr_input[restart_step]',
            'verbosity': 'lr_input[lr_verbosity]',
            'disk_io': 'lr_input[disk_io]',
            'prefix': 'lr_input[prefix]',
            'wfcdir': 'lr_input[wfcdir]',
            'outdir': 'lr_input[outdir]',
            'max_seconds': 'lr_input[max_seconds'
        },
        'lr_control': {
            'itermax': 'lr_control[itermax]',
            'n_pol': 'lr_control[n_ipol]',
            'ipol': 'lr_control[ipol]',
            'ltammd': ('lr_control[ltammd]', options.set_boolean_flag, None),
            'lrpa': ('lr_control[lrpa]', options.set_boolean_flag, None),
            'charge_response': ('lr_control[charge_response]', options.set_boolean_flag, None),
            'tqr': ('lr_control[tqr]', options.set_boolean_flag, None),
            'auto_rs': ('lr_control[auto_rs]', options.set_boolean_flag, None),
            'no_hxc': ('lr_control[no_hxc]', options.set_boolean_flag, None),
            'lproject': ('lr_control[lproject]', options.set_boolean_flag, None),
            'scissor': 'lr_control[scissor]',
            'ecutfock': 'lr_control[ecutfock]',
            'pseudo_hermitian': ('lr_control[pseudo_hermitian]', options.set_boolean_flag, None),
            'd0psi_rs': ('lr_control[d0psi_rs]', options.set_boolean_flag, None),
            'lshift_d0psi': ('lr_control[d0psi_rs]', options.set_boolean_flag, None),
            'q1': 'lr_control[q1]',
            'q2': 'lr_control[q2]',
            'q3': 'lr_control[q3]',
            'eels_approx': 'lr_control[approximation]'
        },
        'lr_davidson': {
            'num_eignv': 'lr_dav[num_eign]',
            'num_init': 'lr_dav[num_init]',
            'num_basis_max': 'lr_dav[num_basis_max]',
            'res_conv_thr': 'lr_dav[res_conv_thr]',
            'precondition': ('lr_dav[precondition]', options.set_boolean_flag, None),
            'reference': 'lr_dav[reference]',
            'single_pole': ('lr_dav[single_pole]', options.set_boolean_flag, None),
            'sort_contr': ('lr_dav[sort_contr', options.set_boolean_flag, None),
            'diag_of_h': ('lr_dav[diag_of_h]', options.set_boolean_flag, None),
            'close_pre': ('lr_dav[close_pre]', options.set_boolean_flag, None),
            'broadening': 'lr_dav[broadening]',
            'print_spectrum': ('lr_dav[print_spectrum]', options.set_boolean_flag, None),
            'start': 'lr_dav[start]',
            'finish': 'lr_dav[finish]',
            'step': 'lr_dav[step]',
            'if_checkorth': ('lr_dav[if_checkorth]', options.set_boolean_flag, None),
            'if_random_init': ('lr_dav[if_random_init]', options.set_boolean_flag, None),
            'if_check_her': ('lr_dav[if_check_her]', options.set_boolean_flag, None),
            'p_nbnd_occ': 'lr_dav[p_nbnd_occ]',
            'p_nbnd_virt': 'lr_dav[p_nbnd_virt]',
            'poor_of_ram': ('lr_dav[poor_of_ram]', options.set_boolean_flag, None),
            'max_iter': 'lr_dav[max_iter]',
            'ecutfock': 'lr_dav[ecutfock]',
            'conv_assistant': ('lr_dav[conv_assistant]', options.set_boolean_flag, None),
            'if_dft_spectrum': ('lr_dav[if_dft_spectrum]', options.set_boolean_flag, None),
            'no_hxc': ('lr_dav[no_hxc]', options.set_boolean_flag, None),
            'd0psi_rs': ('lr_dav[d0psi_rs]', options.set_boolean_flag, None),
            'lshift_d0psi': ('lr_dav[lshift_d0psi]', options.set_boolean_flag, None),
            'lplot_drho': ('lr_dav[lplot_drho]', options.set_boolean_flag, None),
            'vccouple_shift': 'lr_dav[vccouple_shift]',
            'ltammd': ('lr_dav[ltammd]', options.set_boolean_flag, None)
        },
        'lr_post': {
            'omeg': 'lr_post[omeg]',
            'beta_z_gamma_prefix': 'lr_post[beta_z_gamma_prefix]',
            'w_T_npol': 'lr_post[w_T_npol]',
            'plot_type': 'lr_post[plot_type]',
            'epsil': 'lr_post[epsil]',
            'iter_maxinit': 'lr_post[iter_maxinit]',
            'sum_rule': ('lr_post[sum_rule]', options.set_boolean_flag, None)
        }
    }

    def __init__(self, **_kwargs):
        super(TdInputConverter, self).__init__(
            *conversion_maps_builder(self.TD_TEMPLATE_MAP),
            input_namelists=('cache', 'lr_input', 'lr_control', 'lr_dav', 'lr_post')
        )

    def get_qe_input(self):
        """
        Overrides superclass get_qe_input with use_defaults set to False.
        :return: the input as obtained from its input builder
        """
        temp = super().get_qe_input().split('\n')
        td = temp[1]
        start = temp.index('&lr_input')
        end = start + temp[start:].index('/')
        qe_input = '\n'.join(temp[start:end + 1])
        """ put lr_input in qe_input """

        if td.lower() in ('lanczos', 'eels'):
            start = temp.index('&lr_control')
        elif td.lower() == 'davidson':
            start = temp.index('&lr_dav')
        end = start + temp[start:].index('/')
        qe_input = qe_input + '\n' + '\n'.join(temp[start:end + 1])
        """ put one of lr_control(lanczos) or lr_davidson in qe_input"""

        start = temp.index('&lr_post')
        end = start + temp[start:].index('/')
        if end - start > 1:
            qe_input = qe_input + '\n' + '\n'.join(temp[start:end + 1])
        """ if not empty add lr_post namelist to qe_input """
        return qe_input


class TdSpectrumInputConverter(RawInputConverter):
    """
    Converts the XML input file described by qes_spectrum scheme in
    namelist input for turbo_spectrum post-processing tool.
    """
    SPEC_TEMPLATE_MAP = {
        'itermax': 'lr_input[itermax]',
        'itermax0': 'lr_input[itermax0]',
        'itermax_actual': 'lr_input[itermax_actual]',
        'extrapolation': 'lr_input[extrapolation]',
        'start': 'lr_input[start]',
        'end': 'lr_input[end]',
        'increment': 'lr_input[increment]',
        'ipol': 'lr_input[ipol]',
        'outdir': 'lr_input[outdir]',
        'prefix': 'lr_input[prefix]',
        'epsil': 'lr_input[epsil]',
        'sym_op': 'lr_input[sym_op]',
        'verbosity': 'lr_input[verbosity]',
        'units': 'lr_input[units]',
        'td': 'lr_input[td]',
        'eign_file': 'lr_input[eign_file]',
        'eels': ('lr_input[eels]', options.set_boolean_flag, None),
    }

    def __init__(self, **_kwargs):
        super(TdSpectrumInputConverter, self).__init__(
            *conversion_maps_builder(self.SPEC_TEMPLATE_MAP),
            input_namelists=['lr_input']
        )


TD_spctInConverter = TdSpectrumInputConverter


# XSpectra Input Converter
class XSpectraInputConverter(RawInputConverter):
    """
    Converts the XML input file described by XSPECTRA scheme in
    namelist input for XSPECTRA post-processing tool.
    """
    XSPECTRA_TEMPLATE_MAP = {
        'input_xspectra': {'calculation': 'input_xspectra[calculation]',
                           'verbosity': 'input_xspectra[verbosity]',
                           'prefix': 'input_xspectra[prefix]',
                           'outdir': 'input_xspectra[outdir]',
                           'xiabs': 'input_xspectra[xiabs]',
                           # xkvec
                           'xkvec1': ('input_xspectra[xkvec1]', options.get_xspectra_c, None),
                           'xkvec2': ('input_xspectra[xkvec2]', options.get_xspectra_c, None),
                           'xkvec3': ('input_xspectra[xkvec3]', options.get_xspectra_c, None),
                           # xepsilon
                           'xepsilon1': ('input_xspectra[xepsilon1]', options.get_xspectra_c, None),
                           'xepsilon2': ('input_xspectra[xepsilon2]', options.get_xspectra_c, None),
                           'xepsilon3': ('input_xspectra[xepsilon3]', options.get_xspectra_c, None),
                           'xcoordcrys': 'input_xspectra[xcoordcrys]',
                           'ef_r': 'input_xspectra[ef_r]',
                           'xe0': 'input_xspectra[xe0]',
                           'xonly_pot': 'input_xspectra[xonly_pot]',
                           'xread_wf': 'input_xspectra[xread_wf]',
                           'x_save_file': 'input_xspectra[x_save_file]',
                           'xniter': 'input_xspectra[xniter]',
                           'xerror': 'input_xspectra[xerror]',
                           'xcheck_conv': 'input_xspectra[xcheck_conv]',
                           'show_status': 'input_xspectra[show_status]',
                           'nelup': 'input_xspectra[nelup]',
                           'neldw': 'input_xspectra[neldw]',
                           'U_projection_type': 'input_xspectra[U_projection_type]',
                           'time_limit': 'input_xspectra[time_limit]',
                           'restart_mode': 'input_xspectra[restart_mode]',
                           'edge': 'input_xspectra[edge]',
                           'lplus': 'input_xspectra[lplus]',
                           'lminus': 'input_xspectra[lminus]'},
        'plot': {'xnepoint': 'plot[xnepoint]',
                 'xgamma': 'plot[xgamma]',
                 'xemax': 'plot[xemax]',
                 'xemin': 'plot[xemin]',
                 'cut_occ_states': 'plot[cut_occ_states]',
                 'terminator': 'plot[terminator]',
                 'gamma_mode': 'plot[gamma_mode]',
                 'gamma_file': 'plot[gamma_file]',
                 'gamma_energy': 'plot[gamma_energy]',
                 'gamma_value': 'plot[gamma_value]',
                 'xanes_file': 'plot[xanes_file]'},
        'pseudos': {'filecore': 'pseudos[filecore]',
                    'filerecon': 'pseudos[filerecon]',
                    # rpaw
                    'r_paw0': ('pseudos[r_paw0]', options.get_xspectra_c, None),
                    'r_paw1': ('pseudos[r_paw1]', options.get_xspectra_c, None),
                    'r_paw2': ('pseudos[r_paw2]', options.get_xspectra_c, None),
                    'r_paw3': ('pseudos[r_paw3]', options.get_xspectra_c, None)},
        'cut_occ': {'cut_ierror': 'cut_occ[cut_ierror]',
                    'cut_stepu': 'cut_occ[cut_stepu]',
                    'cut_stepl': 'cut_occ[cut_stepl]',
                    'cut_startt': 'cut_occ[cut_startt]',
                    'cut_tinf': 'cut_occ[cut_tinf]',
                    'cut_tsup': 'cut_occ[cut_tsup]',
                    'cut_desmooth': 'cut_occ[cut_desmooth]',
                    'cut_nmemu': 'cut_occ[cut_nmemu]',
                    'cut_nmeml': 'cut_occ[cut_nmeml]'},
        'k_points_IBZ': ('K_POINTS', cards.get_xspectra_k_points_card, None),
    }

    def __init__(self, **_kwargs):
        super(XSpectraInputConverter, self).__init__(
            *conversion_maps_builder(self.XSPECTRA_TEMPLATE_MAP),
            input_namelists=('input_xspectra', 'plot', 'pseudos', 'cut_occ'),
            input_cards=("K_POINTS",)
        )


class EPWInputConverter(RawInputConverter):
    """
    converts the XML input file for EPW to the namelist format
    expected by epw.x
    """
    EPW_TEMPLATE_MAP = {
        'control_variables': {
            'prefix': "inputepw[prefix]",
            'outdir': "inputepw[outdir]",
            'iverbosity': "inputepw[iverbosity]",
            'dvscf_dir': "inputepw[dvscf_dir]",
            'filukk': "inputepw[filukk]",
            'elph': "inputepw[elph]",
            'ep_coupling': "inputepw[ep_coupling]",
            'elecselfen': "inputepw[elecselfen]",
            'phonselfen': "inputepw[phonselfen]",
            'lindabs': "inputepw[lindabs]",
            'band_plot': "inputepw[band_plot]",
            'fermi_plot': "inputepw[fermi_plot]",
            'cumulant': "inputepw[cumulant]",
            'prtgkk':  "inputepw[prtgkk]",
            'epbread': "inputepw[epbread]",
            'epbwrite': "inputepw[epbwrite]",
            'epwread': "inputepw[epwread]",
            'epwwrite': "inputepw[epwwrite]",
            'ephwrite': "inputepw[ephwrite]",
            'eig_read': "inputepw[eig_read]",
            'delta_approx': "inputepw[delta_approx]",
            'eps_acustic': "inputepw[eps_acustic]",
            'etf_mem': "inputepw[etcmem]",
            'nbndsub': "inputepw[nbndsub]",
            'nq1': "inputepw[nq1]",
            'nq2': "inputepw[nq2]",
            'nq3': "inputepw[nq3]",
            'nk1': "inputepw[nk1]",
            'nk2': "inputepw[nk2]",
            'nk3': "inputepw[nk3]",
            'nqf1': "inputepw[nqf1]",
            'nqf2': "inputepw[nqf2]",
            'nqf3': "inputepw[nqf3]",
            'nkf1': "inputepw[nkf1]",
            'nkf2': "inputepw[nkf2]",
            'nkf3': "inputepw[nkf3]",
            'mp_mesh_k': "inputepw[mp_mesh_k]",
            'filqf': "inputepw[filqf]",
            'filkf': "inputepw[filkf]",
            'vme': "inputepw[vme]",
            'degaussw': "inputepw[degaussw]",
            'degaussq': "inputepw[degaussq]",
            'fsthick': "inputepw[fsthick]",
            'ngaussw': "inputepw[ngaussw]",
            'nsmear': "inputepw[nsmear]",
            'delta_smear': "inputepw[delta_smear]",
            'restart': "inputepw[restart]",
            'restart_step': "inputepw[restart_step]",
            'scissor': "inputepw[scissor]",
            'lphase': "inputepw[lphase]",
            'lpolar': "inputepw[lpolar]",
            'efermi_read': "inputepw[efermi_read]",
            'fermi_energy': "inputepw[fermi_energy]",
            'lscreen': "inputepw[lscreen]",
            'scr_typ': "inputepw[scr_typ]",
            'fermi_diff': "inputepw[fermi_diff]",
            'smear_rpa': "inputepw[smear_rpa]",
            'lifc': "inputepw[lifc]",
            'asr_typ': "inputepw[asr_typ]",
            'wannierize': "inputepw[wannierize]",
            'amass': {'$': ("inputepw[amass]", options.set_one_amass_line, None)}
            },
        'wannier90': {
            'num_iter': "inputepw[num_iter]",
            'dis_win_max': "inputepw[dis_win_max]",
            'dis_win_min': "inputepw[dis_win_min]",
            'dis_froz_min': "inputepw[dis_froz_min]",
            'dis_froz_max': "inputepw[dis_frox_max]",
            'proj': {'$': ("inputepw[proj]", options.set_one_proj_line, None)},
            'bands_skipped': "inputepw[bands_skipped]",
            'iprint': "inputepw[iprint]",
            'wannier_plot': ["inputepw[wannier_plot]",
                             ("inputepw[wdata]", options.set_wdata_lines, None)
                             ],
            'wannier_plot_supercell': "inputepw[wannier_plot_supercell]",
            'wannier_plot_scale': "inputepw[wannier_plot_scale]",
            'wannier_plot_radius': "inputepw[wannier_plot_radius]",
            'wannier_plot_list': {'@segment': ("inputepw[wdata]",
                                               options.set_wdata_lines, None)},
            'wannier_plot_format': ('inputepw[wdata]',
                                    options.set_wdata_lines, None),
            'use_ws': ("inputepw[wdata]", options.set_wdata_lines, None),
            'reduce_unk': "inputepw[reduce_unk]",
            'scdm_proj': "inputepw[scdm_proj]",
            'scdm_sigma': "inputepw[scdm_sigma]",
            'auto_projections': "inputepw[auto_projections]",
            'scdm_entanglement': "inputepw[scdm_entaglement]",
            'scdm_mu': "inputepw[scdm_mu]",
        }
    }

    def __init__(self, **kwargs):
        super().__init__(*conversion_maps_builder(self.EPW_TEMPLATE_MAP),
                         input_namelists=['inputepw'],
                         input_cards=[])



#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato
#
"""
reads xml inputs and suggest smart parallelization paremeters
"""
import sys

from pyrsistent import b
import qeschema
from qeschema.utils import etree_iter_path
 


def parse_args():
   """
   command arguent parsing
   """
   import argparse
   parser=argparse.ArgumentParser(
      description="This program reads an XML input and some information about the node"
                  "and prints a suggestion of smart parallelizazion parameters"
                  )
   parser.add_argument("-v", "--verbosity", action="count", default=1,
                        help="Increase output verbosity.")
   parser.add_argument('-in', metavar='FILE', required=True, help="XML input filename.")
   parser.add_argument('-nodes', metavar='INTEGER', required=False, default=1, 
                        help="Number of nodes")
   parser.add_argument('-cpus_per_node', metavar='INTEGER', required=False, default=1,
                     help="CPUS per node" )
   parser.add_argument('-mem', metavar="GigaBytes", required=True, help="Memory per Node in GB" )
   parser.add_argument('-pseudo_dir', metavar='PATH', required=False, default='./',help="PATH to pseudos")
   return parser.parse_args()

def get_doc_path(doc, path):
      if doc.schema.find(path):
         return doc.schema.find(path).decode(doc.find(path))
      else:
         return None
      
def count_kb(sp, pseudo_dir, positions):
   """
   :sp: dict with species infos
   :pseudo_dir: path to pseudos
   :positions: dict with atomic positions infos
   counts KB projectors in a calculation 
   :return: int  the number of KB used
   """
   from pathlib import Path
   from xml.etree import ElementTree as Etree
   p = Path(pseudo_dir, sp['pseudo_file'])
   tree = Etree.parse(p)
   betas = [_ for _ in tree.findall('.//PP_NONLOCAL/*') if 'PP_BETA' in _.tag]
   kb = sum((2 * int(dict( _.items())['angular_momentum']) + 1 for _ in betas)) 
   count_sp = sum((1 for _ in positions['atom'] if _['@name'] == sp['@name'] ))
   return count_sp*len(betas), count_sp*kb 

def estimate_grid(basis, cell):

   from math import sqrt
   a1 = cell['a1']
   a2 = cell['a2']
   a3 = cell['a3']
   if basis.get('ecutrho'):
      gcut = sqrt(float(basis.get('ecutrho')))
   else:
      gcut= 2.0 * sqrt(float.basis.get('ecutwfc'))
   
   import pdb
   pdb.set_trace()
   nr1 = int(round(gcut * sqrt(sum((x*x for x in a1))),0))
   nr2 = int(round(gcut * sqrt(sum((x*x for x in a2))),0)) 
   nr3 = int(round(gcut * sqrt(sum((x*x for x in a3))),0)) 
   return nr1, nr2, nr3 

if __name__=='__main__' :
   if sys.version_info < (3, 5, 0):
      sys.stderr.write("You need python 3.5 or later to run this program\n")
      sys.exit(1)

   args = parse_args()
   print("Smart parametrization parameters for  %r ...\n" % getattr(args, 'in'))

   if __package__ is None:
      from os import path
      sys.path.append(path.abspath(path.dirname(__file__)+'/../'))

   import qeschema
   import os
   import xml.etree.ElementTree as Etree 
   args = parse_args()
   input_fn = getattr(args,'in')
   pseudo_dir = getattr(args, 'pseudo_dir')  
   tree = Etree.parse(input_fn)
   root = tree.getroot() 
   element_name = root.tag.split('}')[-1]
   if element_name == 'espresso':
      doc = qeschema.PwDocument(source=input_fn)
   else: 
      sys.stderr.write("Could not find correct XML in %s, exiting...\n" % input_fn)
      sys.exit(1)
   root = None
   tree = None

   basis = get_doc_path(doc,'.//input//basis')
   cell  = get_doc_path(doc, './/input//cell')
   species = get_doc_path(doc, './/input//atomic_species')['species']
   try:
      positions = get_doc_path(doc, './/input//atomic_positions')
   except AttributeError:
      positions = get_doc_path(doc, './/input//crystal_positions')
   import pdb
   nkb   = sum((count_kb(_,pseudo_dir, positions)[0] for _ in species))
   nwfcat =  sum((count_kb(_,pseudo_dir, positions)[1] for _ in species)) 
   print (f"nkb={nkb}, n_wfc_at={nwfcat}")
   fft_grid = estimate_grid(basis, cell)
   print ("estimated fff grid = ("+":".join(["estimated fft grid (",]+[f"{_}" for _ in fft_grid])+")")  

         
      



                                                      


#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato
#

###
# XML Schema namespaces

XSD_NAMESPACE = 'http://www.w3.org/2001/XMLSchema'
"URI of the XML Schema Definition namespace (xs|xsd)"

XSI_NAMESPACE = 'http://www.w3.org/2001/XMLSchema-instance'
"URI of the XML Schema Instance namespace (xsi)"

XML_NAMESPACE = 'http://www.w3.org/XML/1998/namespace'
"URI of the XML namespace (xml)"

###
# QE namespaces

PW_NAMESPACE = "http://www.quantum-espresso.org/ns/qes/qes-1.0"
"URI of the Quantum Espresso namespace for PW application"

PHONON_NAMESPACE = "http://www.quantum-espresso.org/ns/qes/qes_ph_1.0"
"URI of the Quantum Espresso namespace for Phonon application"

NEB_NAMESPACE = "http://www.quantum-espresso.org/ns/neb"
"URI of the Quantum Espresso namespace for NEB application"

TD_NAMESPACE = "http://www.quantum-espresso.org/ns/neb"
"URI of the Quantum Espresso namespace for TDDFPT application"

TD_SPECTRUM_NAMESPACE = "http://www.quantum-espresso.org/ns/qes/qes_spectrum-1.0"
"URI of the Quantum Espresso namespace for turbo-spectrum application"

# XSpectra namespace
XSPECTRA_NAMESPACE = "http://www.quantum-espresso.org/ns/qes/qes_xspectra-1.0"
"URI of the Quantum Espresso namespace for XSPECTRA application"



#
# Copyright (c), 2015-2020, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# Authors: Davide Brunato, Giovanni Borghi
#
"""
Conversion functions for Quantum Espresso cards.
"""
import logging

logger = logging.getLogger('qeschema')


#
# Functions for QE cards
#
def get_atomic_species_card(name, **kwargs):
    """
    Convert XML data to ATOMIC_SPECIES card

    :param name: Card name
    :param kwargs: Dictionary with converted data from XML file
    :return: List of strings
    """
    try:
        atomic_species = kwargs['atomic_species']
        species = atomic_species['species']
    except KeyError as err:
        logger.error("Missing required arguments when building ATOMIC_SPECIES card! %s" % err)
        return []

    lines = [name]
    try:
        lines.append(
            ' {0} {1} {2}'.format(species['@name'], species['mass'], species['pseudo_file'])
        )
    except TypeError:
        for specie in species:
            lines.append(
                ' {0} {1} {2}'.format(specie['@name'], specie['mass'], specie['pseudo_file'])
            )
    return lines


def get_positions_units(item):
    """
    Read different type of positions and return them together with units.

    :param item: Data item
    :return: Atomic positions dict and unit, empty dict and None if not found
    """

    ptypes = ('atomic_positions', 'crystal_positions', 'wyckoff_positions')
    units = ('bohr', 'crystal', 'crystal_sg')
    for ptype, unit in zip(ptypes, units):
        if ptype in item:
            return item[ptype], unit

    return {}, None


def get_atomic_positions_cell_card(name, **kwargs):
    """
    Convert XML data to ATOMIC_POSITIONS card

    :param name: Card name
    :param kwargs: Dictionary with converted data from XML file
    :return: List of strings
    """
    try:
        atomic_structure = kwargs['atomic_structure']
    except KeyError:
        logger.error("Missing required arguments when building ATOMIC_POSITIONS card!")
        return []

    # Find atoms
    positions, units = get_positions_units(atomic_structure)
    try:
        atoms = positions['atom']
    except KeyError:
        logger.error("Cannot find any atoms for building ATOMIC_POSITIONS!")
        return []

    if not isinstance(atoms, list):
        atoms = [atoms]

    # Check atoms with position constraints
    free_positions = kwargs.get('free_positions')
    if free_positions:
        free_positions = free_positions.get('$', [])

    if free_positions and len(free_positions) != 3 * len(atoms):
        logger.error("ATOMIC_POSITIONS: incorrect number of position constraints!")

    # Add atomic positions
    lines = ['%s %s' % (name, units)]
    for k in range(len(atoms)):
        try:
            line = '{:4}'.format(atoms[k].get('@name'))
            line += ' {:12.8f}  {:12.8f}  {:12.8f}'.format(*atoms[k].get('$'))
        except (ValueError, TypeError):
            logger.error("ATOMIC_POSITIONS: incorrect datatype in positions!")
            continue

        if free_positions:
            try:
                if free_positions[3 * k] + \
                        free_positions[3 * k + 1] + \
                        free_positions[3 * k + 2] != 3:
                    line += ' {:4d}{:4d}{:4d}'.format(*map(int, free_positions[3 * k:3 * k + 3]))
            except IndexError:
                pass

        lines.append(line)

    return lines


def get_atomic_constraints_card(name, **kwargs):
    """
    Convert XML data to CONSTRAINTS card

    :param name: Card name
    :param kwargs: Dictionary with converted data from XML file
    :return: List of strings
    """
    try:
        num_of_constraints = kwargs['atomic_constraints']['num_of_constraints']
        tolerance = kwargs['atomic_constraints']['tolerance']
        atomic_constraints = kwargs['atomic_constraints']['atomic_constraint']
    except KeyError:
        logger.error("Missing required arguments when building CONSTRAINTS card!")
        return []

    lines = [name, f"{num_of_constraints} {tolerance}"]
    for constraint in atomic_constraints:
        constr_parms = constraint['constr_parms']  # list with at most 4 float items
        constr_type = constraint['constr_type']  # string
        constr_target = constraint.get('constr_target', "")  # float or empty
        lines.append(f"{constr_type} {' '.join([str(_) for _ in constr_parms])} {constr_target}")
    return lines


def get_k_points_card(name, **kwargs):
    """
    Convert XML data to K_POINTS card

    :param name: Card name
    :param kwargs: Dictionary with converted data from XML file
    :return: List of strings
    """
    k_point = {}  # type: Union[dict, List[dict]]
    nk = monkhorst_pack = None
    gamma_only = kwargs.get('gamma_only', False)
    try:
        if not gamma_only:
            k_points_ibz = kwargs['k_points_IBZ']
            monkhorst_pack = k_points_ibz.get('monkhorst_pack', {})
            if monkhorst_pack:
                k_attrib = 'automatic'
            else:
                k_attrib = None
                k_point = k_points_ibz['k_point']
                nk = k_points_ibz['nk']
        else:
            k_attrib = 'gamma'

    except KeyError as err:
        logger.error("Missing required arguments when building K_POINTS card! %s" % err)
        return []
    else:
        if not isinstance(k_point, list):
            k_point = [k_point]

    lines = [name] if k_attrib is None else ['%s %s' % (name, k_attrib)]
    if k_attrib is None:
        lines.append(' {}'.format(nk))
        for point in k_point:
            lines.append(' {0} {1}'.format(
                ' '.join([str(value) for value in point['$']]), point['@weight'])
            )
    elif k_attrib == 'automatic':
        lines.append(' %(@nk1)s %(@nk2)s %(@nk3)s %(@k1)s %(@k2)s %(@k3)s' % monkhorst_pack)

    return lines


def get_hubbard_card(name, **kwargs):
    """
    writes the hubbard card for new format
    """
    try:
        new_format = kwargs['dftU']['@new_format']
    except KeyError:
        new_format = False
    if not new_format:
        return []
    try:
        dftu = kwargs['dftU']
    except KeyError as err:
        logger.error("Missing required argument %s when building "
                     "parameter %r", str(err), name)
        return []

    projtype = kwargs.get('U_projection_type', 'atomic')
    lines = [f"{name} {projtype}"]
    for tag in iter(['U', 'J0', 'alpha', 'beta']):
        lines.extend(_hubbard_lines(dftu, tag))
    for tag in iter(['J', 'U2', 'V']):
        lines.extend(_hubbard_special_lines(dftu, tag))
    return lines


def _hubbard_lines(dftu, tag):
    related_data = dftu.get(f"Hubbard_{tag}", [])
    lines = []
    for value in iter(related_data if isinstance(related_data, list) else [related_data]):
        specie = value['@specie']
        label = value['@label']
        if label != 'no Hubbard':
            lines.append(f"{tag}  {specie}-{label}  {value['$']:8.3f}")
    return lines


def _hubbard_special_lines(dftu, tag):
    related_data = dftu.get(f"Hubbard_{tag}", [])
    llabels = ['s', 'p', 'd', 'f']
    lines = []
    for value in iter(related_data if isinstance(related_data, list) else [related_data]):
        specie = value['@specie']
        label = value.get('@label', 'not found')
        if label != 'no Hubbard':
            if tag == 'J':
                lines.append(f"J {specie}-{label} {value['$'][0]:8.3f}")
                if 'd' in label:
                    lines.append(f"B {specie}-{label} {value['$'][1]:8.3f}")
                elif 'f' in label:
                    lines.extend([f"E2 {specie}-{label} {value['$'][1]:8.3f}",
                                  f"E3 {specie}-{label} {value['$'][2]:8.3f}"])
            elif tag == 'U2':
                background = value['@background']
                if label == 'not found':
                    def labnl(nnum, lnum):
                        return f"{nnum}{llabels[lnum-1]}"

                    label = labnl(value['n2_number'], value['l2_number'])
                    if 'two' in background:
                        label = f"{label}-{labnl(value['n3_number'],value['l3_number'])}"
                    lines.append(f"U {specie}-{label}  {value['$']:8.3f}")
            elif tag == 'V':
                speclab1 = f"{value['@specie1']}-{value['@label1']}"
                speclab2 = f"{value['@specie2']}-{value['@label2']}"
                index1 = value['@index1']
                index2 = value['@index2']
                lines.append(f"{tag} {speclab1} {speclab2} {index1} {index2} {value['$']:8.3f}")
    return lines


def get_atomic_forces_card(name, **kwargs):
    """
    Convert XML data to ATOMIC_FORCES card

    :param name: Card name
    :param kwargs: Dictionary with converted data from XML file
    :return: List of strings
    """
    try:
        external_atomic_forces = kwargs['external_atomic_forces']
    except KeyError:
        logger.debug("Missing required arguments when building ATOMIC_FORCES card!")
        return []

    # Warning if number of atoms in atomic positions differ with forces
    atomic_positions = kwargs.get('atomic_positions', {})
    atoms = atomic_positions.get('atom', [])
    if atoms and len(atoms) != len(external_atomic_forces) / 3:
        logger.error("incorrect number of atomic forces")

    # Build input card text lines
    lines = [name]
    f = external_atomic_forces
    for at in atoms:
        try:
            line = '{:<3s}  {:14.8f} {:14.8f} {:14.8f}'.format(at["@name"], f[0], f[1], f[2])
        except IndexError:
            pass
        else:
            lines.append(line)
            f = f[3:]
    return lines


def _get_cell_lines(name, cells):
    """
    Return cell lines

    :param str name: Card name
    :params dict cells: Lattice params
    :return list[str]: List of strings representing the CELL
    """

    if cells:
        lines = ['%s bohr' % name]
        for key in sorted(cells):
            if key not in ['a1', 'a2', 'a3']:
                continue

            lines.append((3 * '{:12.8f} ').format(*cells[key]))
        return lines

    return []


def get_cell_parameters_card(name, **kwargs):
    """
    Convert XML data to CELL_PARAMETERS card

    :param name: Card name
    :param kwargs: Dictionary with converted data from XML file
    :return: List of strings
    """
    try:
        atomic_structure = kwargs['atomic_structure']
    except KeyError:
        logger.error("Missing required arguments when building CELL_PARAMETERS card!")
        return []
    # Add cell parameters card
    cells = atomic_structure.get('cell', {})
    return _get_cell_lines(name, cells)


#
# Phonon Cards
#
def get_qpoints_card(name, **kwargs):
    assert isinstance(name, str)
    try:
        ldisp = kwargs['ldisp']
    except KeyError:
        ldisp = False
    else:
        if ldisp:
            return []

    qplot = kwargs.get('qplot', False)
    if not qplot and not ldisp:
        try:
            xq = kwargs['xq']
        except KeyError:
            xq = [0.e0, 0.e0, 0.e0]
        line = "{:6.4f}  {:8.4f}  {:8.4f}".format(xq[0], xq[1], xq[2])
        return [line]

    lines = []
    try:
        nqs = kwargs['nqs']
    except KeyError:
        nqs = 1
        logger.warning("qplot was set to true in input but no value "
                       "for nqs was provided: assuming nqs=1")

    lines.append('{:4d}'.format(nqs))
    q_points_list = kwargs['q_points_list']['q_point']
    for q_point in q_points_list:
        vector = ' '.join([str(coord) for coord in q_point['$']])
        lines.append(' %s %s' % (vector, q_point['@weight']))
    return lines


def get_nat_todo_card(name, **kwargs):

    natom = kwargs['nat_todo']['@natom']
    if natom == 0:
        assert 'atom' not in kwargs['nat_todo']
        return []

    atoms = kwargs['nat_todo']['atom']

    assert natom == len(atoms)
    return [' '.join(map(str, atoms))]


def get_climbing_images(name, **kwargs):
    assert isinstance(name, str)
    try:
        manual_images = kwargs['climbingImage'] == 'manual' or kwargs['climbingImage'] == 'MANUAL'
    except KeyError:
        manual_images = False

    if not manual_images:
        return ['']

    if isinstance(kwargs['climbingImageIndex'], list):
        line = [int(x) for x in kwargs['climbingImageIndex']]
        fmt = len(line) * ' %d, '
        line = fmt % tuple(line)
    else:
        line = ' %d ' % int(kwargs['climbingImageIndex'])
    return [line]


def get_neb_images_positions_card(name, **kwargs):
    """
    Extract atomic positions for each image provided in engine with an atomic_structure element

    :param name: Card name
    :param kwargs: List of dictionaries each containing an atomic_structure element.
    :return: List of lines
    """
    assert isinstance(name, str)
    images = kwargs.get('atomic_structure', [])
    try:
        assert isinstance(images, list)
    except AssertionError:
        images = [images]

    if len(images) < 2:
        logger.error("at least the atomic structures for first and last image should be provided")
        return []

    free_positions = kwargs.get('free_positions', None)
    if free_positions:
        free_positions = free_positions.get('$')
    else:
        free_positions = []

    first_nat = 0
    lines = ['BEGIN_POSITIONS ', 'FIRST_IMAGE ']

    for pos, item in enumerate(images):
        positions, units = get_positions_units(item)
        atoms = positions.get('atom', [])

        if pos == 0:
            first_nat = len(atoms)
            if first_nat <= 0:
                logger.error("no atomic coordinates provided for first image")
                return ''
            if first_nat != int(item.get('@nat', 0)):
                logger.error("nat provided in first image differs from number "
                             "of atoms in atomic_positions!!!")
                return ''
            if free_positions and len(free_positions) != 3 * first_nat:
                logger.error("ATOMIC_POSITIONS: incorrect number of position constraints!")
                return ''
        else:
            if len(atoms) != first_nat:
                logger.error('found images with differing number of atoms !!!')
                return ''

            if pos < len(images) - 1:
                lines.append('INTERMEDIATE_IMAGE ')
            else:
                lines.append('LAST_IMAGE ')

        lines.append('ATOMIC_POSITIONS { %s }' % units)

        # reshape to 3x3
        free_positions_sq = [free_positions[i:i + 3]
                             for i in range(0, len(free_positions), 3)]

        for k, atom in enumerate(atoms):
            sp_name = '{:4}'.format(atom['@name'])
            coords = '{:12.8f}  {:12.8f}  {:12.8f}'.format(*atom['$'])

            if free_positions_sq and sum(free_positions_sq[k]) < 3:
                free_pos = '{:4d}{:4d}{:4d}'.format(*free_positions_sq[k])
                lines.append('%s %s %s' % (sp_name, coords, free_pos))
            else:
                lines.append('%s %s' % (sp_name, coords))

    lines.append('END_POSITIONS ')
    return lines


def get_neb_cell_parameters_card(name, **kwargs):
    """
    Extract cell parameter from the first of the atomic_structure elements provided in engine

    :param name: card name
    :param kwargs: list of the atomic_structure dictionaries translate for xml element engine
    :return: list of text lines for the cell parameters card
    """
    assert isinstance(name, str)
    images = kwargs.get('atomic_structure', [])
    if not isinstance(images, list):
        images = [images]

    if len(images) < 1:
        logger.error(" No atomic_structure element found in kwargs !!!")
        return []

    atomic_structure = images[0]
    cells = atomic_structure.get('cell', {})
    return _get_cell_lines(name, cells)


def get_neb_atomic_forces_card(name, **kwargs):
    # TODO
    assert isinstance(name, str)
    assert isinstance(kwargs, dict)


def get_xspectra_k_points_card(name, **kwargs):
    """
    Convert XML data to K_POINTS card for xspectra calculation.

    :param name: Card name
    :param kwargs: Dictionary with converted data from XML file
    :return: List of strings
    """
    lines = []
    k_points_ibz = kwargs['k_points_IBZ']
    monkhorst_pack = k_points_ibz.get('monkhorst_pack', {})

    lines.append(' %(@nk1)s %(@nk2)s %(@nk3)s %(@k1)s %(@k2)s %(@k3)s' % monkhorst_pack)

    return lines



#!/usr/bin/env python
#
# Copyright (c), 2015-2022, Quantum Espresso Foundation and SISSA (Scuola
# Internazionale Superiore di Studi Avanzati). All rights reserved.
# This file is distributed under the terms of the MIT License. See the
# file 'LICENSE' in the root directory of the present distribution, or
# http://opensource.org/licenses/MIT.
#
# @author Davide Brunato
#
from setuptools import setup

with open("README.rst") as readme:
    long_description = readme.read()

setup(
    name='qeschema',
    version='1.5.1',
    install_requires=['xmlschema>=1.6.4,<2.4.0', 'numpy'],
    extras_require={
        'HDF5': ['h5py'],
        'YAML': ['pyyaml'],
    },
    packages=['qeschema', 'qeschema.hdf5'],
    package_data={'qeschema': ['schemas/*.xsd', 'schemas/releases/*.xsd']},
    scripts = ['scripts/xml2qeinput.py', 'scripts/yaml2qeinput.py'],
    url='https://github.com/QEF/qeschema',
    authors='Davide Brunato,Pietro Delugas,Giovanni Borghi,Alexandr Fonari',
    license='MIT',
    license_file='LICENSE',
    description='Schema-based tools and interfaces for Quantum Espresso data',
    long_description=long_description,
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Scientific/Engineering :: Physics',
        'Topic :: Utilities',
    ]
)
